# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import tempfile
import importlib
import sys
import os
from openc3.utilities.bucket import Bucket
from openc3.utilities.target_file import TargetFile
from openc3.environment import OPENC3_SCOPE, OPENC3_CONFIG_BUCKET

# The last item in the sys.meta_path is the PathFinder
# PathFinder locates modules according to a path like structure
# which is what we're doing with target file imports
_real_pathfinder = sys.meta_path[-1]
_bucket_client = Bucket.getClient()


class MyLoader(importlib.abc.Loader):
    # Normally this method returns None and Python creates the module
    # However, we're explicitly creating the module and executing it
    # here based on the contents provided by the spec.loader_state.
    def create_module(self, spec):
        if spec.loader_state:
            file = tempfile.NamedTemporaryFile(mode="w+t", suffix=".py")
            file.write(spec.loader_state)
            file.seek(0)  # Rewind so the file is ready to read
            spec = importlib.util.spec_from_file_location(spec.name, file.name)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            file.close()
            return module
        return None

    # Normally this is where the module is executed and populated.
    # However, this let to some recursion issues so we do that above
    # in the create_module method.
    def exec_module(self, module):
        pass


class MyFinder(type(_real_pathfinder)):
    @classmethod
    def find_spec(cls, name, path, target=None):
        spec = _real_pathfinder.find_spec(name, path, target)
        if spec is not None:
            return spec
        # If that doesn't work see if it is a target file
        # We're relying on the fact that COSMOS target names are always
        # capitalized and Python package names are (almost) always lowercase.
        parts = name.split(".")
        if spec is None and parts[0].upper() == parts[0]:
            spec = cls.find_target_file(name)
        # Returning None means we didn't find the spec which
        # triggers Python to raise ModuleNotFoundError
        return spec

    @classmethod
    def find_target_file(cls, name):
        # Convert the import statement: import INST.lib.filename
        # into a path: INST/lib/filename
        path_name = name.replace(".", "/")
        # Bucket paths can not start with '/' and must end in '/'
        path = f"{OPENC3_SCOPE}/targets/{path_name}/"

        # See if anything exists at this path
        dirs, files = _bucket_client.list_files(OPENC3_CONFIG_BUCKET, path)
        if dirs or files:
            # Create a ModuleSpec using our loader
            spec = importlib.machinery.ModuleSpec(name, MyLoader(), origin=None)
            # Must set submodule_search_locations to indicate this is a package
            spec.submodule_search_locations = [name]
            spec.has_location = False

            # Found a path so need to return the spec to make a module
            return spec
        else:
            # Create a ModuleSpec using our loader
            spec = importlib.machinery.ModuleSpec(name, MyLoader(), origin=None)
            # Must set submodule_search_locations to indicate this is a package
            spec.submodule_search_locations = [name]
            spec.has_location = False

            # Try to read the filename based on the last bit of the path
            # NOTE: This handles the target_modified vs target directory
            body = TargetFile.body(OPENC3_SCOPE, f"{path_name}.py")
            if body is None:
                body = TargetFile.body(OPENC3_SCOPE, os.path.join(path_name, "__init__.py"))
            if body is not None:
                # If a file was actually there we assign it to loader_state
                # so the loader can parse it into a real module
                spec.loader_state = body.decode()
                return spec

            # File not found and we are done
            return None


sys.meta_path[-1] = MyFinder



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.


class LimitsResponse:
    """This class defines a #call method which is called case a PacketItem
    goes out of limits. This class must be subclassed and the call method
    implemented. Do NOT use this class directly."""

    # TODO: include Api

    # self.param packet [Packet] Packet the limits response is assigned to
    # self.param item [PacketItem] PacketItem the limits response is assigned to
    # self.param old_limits_state [Symbol] Previous value of the limit. One of None,
    #   'GREEN_HIGH', 'GREEN_LOW', 'YELLOW', 'YELLOW_HIGH', 'YELLOW_LOW',
    #   'RED', 'RED_HIGH', 'RED_LOW'. None if the previous limit state has not yet:
    #   been established.
    def call(self, packet, item, old_limits_state):
        raise RuntimeError("call method must be defined by subclass")

    def to_config(self):
        return f"    LIMITS_RESPONSE {self.__class__.__name__}\n"

    def as_json(self):
        return {"class": self.__class__.__name__}



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.script.suite import Suite, Group
from openc3.script.exceptions import SkipScript


# @deprecated Use SkipScript
class SkipTestCase(SkipScript):
    pass


# @deprecated Use Suite
class TestSuite(Suite):
    def __init__(self):
        super().__init__()
        self.tests = self.scripts
        self.add_test = self.add_group
        self.add_test_case = self.add_script
        self.add_test_setup = self.add_group_setup
        self.add_test_teardown = self.add_group_teardown
        self.run_test = self.run_group
        self.run_test_case = self.run_script
        self.get_num_test = self.get_num_scripts
        self.run_test_setup = self.run_group_setup
        self.run_test_teardown = self.run_group_teardown


# @deprecated Use Group
class Test(Group):
    def __init__(self):
        super().__init__()
        self.run_test_case = self.run_script

    # Alias methods
    test_cases = Group.scripts
    get_num_tests = Group.get_num_scripts
    current_test_suite = Group.current_suite
    current_test_group = Group.current_group
    current_test = Group.current_group
    current_test_case = Group.current_script



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import os
import sys
import time
import json
import threading
import queue
from openc3.microservices.microservice import Microservice
from openc3.system.system import System
from openc3.topics.topic import Topic
from openc3.topics.limits_event_topic import LimitsEventTopic
from openc3.topics.telemetry_decom_topic import TelemetryDecomTopic
from openc3.config.config_parser import ConfigParser
from openc3.utilities.time import to_nsec_from_epoch, from_nsec_from_epoch
from openc3.utilities.thread_manager import ThreadManager
from openc3.microservices.interface_decom_common import (
    handle_build_cmd,
    handle_inject_tlm,
)
from openc3.top_level import kill_thread


class LimitsResponseThread:
    def __init__(self, microservice_name, queue, logger, metric, scope):
        self.microservice_name = microservice_name
        self.queue = queue
        self.logger = logger
        self.metric = metric
        self.scope = scope
        self.count = 0
        self.error_count = 0
        self.metric.set(name="limits_response_total", value=self.count, type="counter")
        self.metric.set(name="limits_response_error_total", value=self.error_count, type="counter")

    def start(self):
        self.thread = threading.Thread(target=self.run, daemon=True)
        self.thread.start()
        ThreadManager.instance().register(self.thread, stop_object=self)
        return self.thread

    def stop(self):
        if self.thread:
            kill_thread(self, self.thread)
            self.thread = None

    def graceful_kill(self):
        self.queue.put([None, None, None])

    def run(self):
        try:
            while True:
                packet, item, old_limits_state = self.queue.get()
                if packet is None:
                    break

                try:
                    item.limits.response.call(packet, item, old_limits_state)
                except Exception as error:
                    self.error_count += 1
                    self.metric.set(name="limits_response_error_total", value=self.error_count, type="counter")
                    self.logger.error(
                        f"{packet.target_name} {packet.packet_name} {item.name} Limits Response Exception!"
                    )
                    self.logger.error(f"Called with old_state = {old_limits_state}, new_state = {item.limits.state}")
                    self.logger.error(repr(error))

                self.count += 1
                self.metric.set(name="limits_response_total", value=self.count, type="counter")
        except Exception as error:
            self.logger.error(f"{self.microservice_name}: Limits Response thread died: {repr(error)}")
            raise error


class DecomMicroservice(Microservice):
    LIMITS_STATE_INDEX = {
        "RED_LOW": 0,
        "YELLOW_LOW": 1,
        "YELLOW_HIGH": 2,
        "RED_HIGH": 3,
        "GREEN_LOW": 4,
        "GREEN_HIGH": 5,
    }

    def __init__(self, *args):
        super().__init__(*args)
        # Should only be one target, but there might be multiple decom microservices for a given target
        # First Decom microservice has no number in the name
        if "__DECOM__" in self.name:
            self.topics.append(f"{self.scope}__DECOMINTERFACE__{{{self.target_names[0]}}}")
        Topic.update_topic_offsets(self.topics)
        System.telemetry.set_limits_change_callback(self.limits_change_callback)
        LimitsEventTopic.sync_system(scope=self.scope)
        self.error_count = 0
        self.metric.set(name="decom_total", value=self.count, type="counter")
        self.metric.set(name="decom_error_total", value=self.error_count, type="counter")
        self.limits_response_queue = queue.Queue()
        self.limits_response_thread = None

    def run(self):
        self.limits_response_thread = LimitsResponseThread(
            microservice_name=self.name,
            queue=self.limits_response_queue,
            logger=self.logger,
            metric=self.metric,
            scope=self.scope,
        )
        self.limits_response_thread.start()

        self.setup_microservice_topic()
        while True:
            if self.cancel_thread:
                break
            try:
                # OpenC3.in_span("read_topics") do
                for topic, msg_id, msg_hash, redis in Topic.read_topics(self.topics):
                    if self.cancel_thread:
                        break

                    if topic == self.microservice_topic:
                        self.microservice_cmd(topic, msg_id, msg_hash, redis)
                    elif "__DECOMINTERFACE__" in topic:
                        if msg_hash.get(b"inject_tlm"):
                            handle_inject_tlm(msg_hash[b"inject_tlm"], self.scope)
                            continue
                        if msg_hash.get(b"build_cmd"):
                            handle_build_cmd(msg_hash[b"build_cmd"], msg_id, self.scope)
                            continue
                    else:
                        self.decom_packet(topic, msg_id, msg_hash, redis)
                        self.metric.set(name="decom_total", value=self.count, type="counter")
                    self.count += 1
                LimitsEventTopic.sync_system_thread_body(scope=self.scope)
            except Exception as error:
                self.error_count += 1
                self.metric.set(name="decom_error_total", value=self.error_count, type="counter")
                self.error = error
                self.logger.error(f"Decom error {repr(error)}")

        self.limits_response_thread.stop()
        self.limits_response_thread = None

    def decom_packet(self, topic, msg_id, msg_hash, _redis):
        # OpenC3.in_span("decom_packet") do
        msgid_seconds_from_epoch = int(msg_id.split("-")[0]) / 1000.0
        delta = time.time() - msgid_seconds_from_epoch
        self.metric.set(
            name="decom_topic_delta_seconds",
            value=delta,
            type="gauge",
            unit="seconds",
            help="Delta time between data written to stream and decom start",
        )

        start = time.time()
        target_name = msg_hash[b"target_name"].decode()
        packet_name = msg_hash[b"packet_name"].decode()

        packet = System.telemetry.packet(target_name, packet_name)
        packet.stored = ConfigParser.handle_true_false(msg_hash[b"stored"].decode())
        # Note: Packet time will be recalculated as part of decom so not setting
        packet.received_time = from_nsec_from_epoch(int(msg_hash[b"received_time"].decode()))
        packet.received_count = int(msg_hash[b"received_count"].decode())
        extra = msg_hash.get(b"extra")
        if extra is not None:
            packet.extra = json.loads(extra)
        packet.buffer = msg_hash[b"buffer"]
        # Processors are user code points which must be rescued
        # so the TelemetryDecomTopic can write the packet
        try:
            packet.process()  # Run processors
        except Exception as error:
            self.error_count += 1
            self.metric.set(name="decom_error_total", value=self.error_count, type="counter")
            self.error = error
            self.logger.error(repr(error))
        # Process all the limits and call the limits_change_callback (as necessary)
        # check_limits also can call user code in the limits response
        # but that is rescued separately in the limits_change_callback
        packet.check_limits(System.limits_set())

        # This is what updates the CVT
        TelemetryDecomTopic.write_packet(packet, scope=self.scope)
        diff = time.time() - start  # seconds as a float
        self.metric.set(name="decom_duration_seconds", value=diff, type="gauge", unit="seconds")

    # Called when an item in any packet changes limits states.
    #
    # @param packet [Packet] Packet which has had an item change limits state
    # @param item [PacketItem] The item which has changed limits state
    # @param old_limits_state [Symbol] The previous state of the item. See
    #   {PacketItemLimits#state}
    # @param value [Object] The current value of the item
    # @param log_change [Boolean] Whether to log this limits change event
    def limits_change_callback(self, packet, item, old_limits_state, value, log_change):
        if self.cancel_thread:
            return
        packet_time = packet.packet_time
        if value:
            message = f"{packet.target_name} {packet.packet_name} {item.name} = {value} is {item.limits.state}"
            if item.limits.values:
                values = item.limits.values[System.limits_set()]
                # Check if the state is RED_LOW, YELLOW_LOW, YELLOW_HIGH, RED_HIGH, GREEN_LOW, GREEN_HIGH
                if DecomMicroservice.LIMITS_STATE_INDEX.get(item.limits.state, None) is not None:
                    # Directly index into the values and return the value
                    message += f" ({values[DecomMicroservice.LIMITS_STATE_INDEX[item.limits.state]]})"
                elif item.limits.state == "GREEN":
                    # If we're green we display the green range (YELLOW_LOW - YELLOW_HIGH)
                    message += f" ({values[1]} to {values[2]})"
                elif item.limits.state == "BLUE":
                    # If we're blue we display the blue range (GREEN_LOW - GREEN_HIGH)
                    message += f" ({values[4]} to {values[5]})"
        else:
            message = f"{packet.target_name} {packet.packet_name} {item.name} is disabled"

        # Include the packet_time in the log json but not the log message
        # Can't use isoformat because it appends "+00:00" instead of "Z"
        time = {"packet_time": packet_time.strftime("%Y-%m-%dT%H:%M:%S.%fZ")}
        if log_change:
            match item.limits.state:
                case "BLUE" | "GREEN" | "GREEN_LOW" | "GREEN_HIGH":
                    # Only print INFO messages if we're changing ... not on initialization
                    if old_limits_state:
                        self.logger.info(message, other=time)
                case "YELLOW" | "YELLOW_LOW" | "YELLOW_HIGH":
                    self.logger.warn(message, other=time, type=self.logger.NOTIFICATION)
                case "RED" | "RED_LOW" | "RED_HIGH":
                    self.logger.error(message, other=time, type=self.logger.ALERT)

        # The openc3_limits_events topic can be listened to for all limits events, it is a continuous stream
        event = {
            "type": "LIMITS_CHANGE",
            "target_name": packet.target_name,
            "packet_name": packet.packet_name,
            "item_name": item.name,
            "old_limits_state": str(old_limits_state),
            "new_limits_state": str(item.limits.state),
            "time_nsec": to_nsec_from_epoch(packet_time),
            "message": str(message),
        }
        LimitsEventTopic.write(event, scope=self.scope)

        if item.limits.response is not None:
            copied_packet = packet.deep_copy()
            copied_item = packet.items[item.name]
            self.limits_response_queue.put([copied_packet, copied_item, old_limits_state])


if os.path.basename(__file__) == os.path.basename(sys.argv[0]):
    DecomMicroservice.class_run()
    ThreadManager.instance().shutdown()
    ThreadManager.instance().join()



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.utilities.logger import Logger


class LimitsParser:
    # self.param parser [ConfigParser] Configuration parser
    # self.param packet [Packet] The current packet
    # self.param cmd_or_tlm [String] Whether this is a command or telemetry packet
    # self.param item [PacketItem] The packet item to create limits on
    # self.param warnings [Array<String>] Array of string warnings which will be
    #   appended with any warnings found case parsing the limits:
    @classmethod
    def parse(cls, parser, packet, cmd_or_tlm, item, warnings):
        if item.states:
            raise parser.error("Items with STATE can't define LIMITS")

        parser = LimitsParser(parser)
        parser.verify_parameters(cmd_or_tlm)
        return parser.create_limits(packet, item, warnings)

    def __init__(self, parser):
        self.parser = parser

    # self.param cmd_or_tlm [String] Whether this is a command or telemetry packet
    def verify_parameters(self, cmd_or_tlm):
        if cmd_or_tlm == "Command":
            raise self.parser.error("LIMITS only applies to telemetry items")

        self.usage = "LIMITS <LIMITS SET> <PERSISTENCE> <ENABLED/DISABLED> <RED LOW LIMIT> <YELLOW LOW LIMIT> <YELLOW HIGH LIMIT> <RED HIGH LIMIT> <GREEN LOW LIMIT (Optional)> <GREEN HIGH LIMIT (Optional)>"
        self.parser.verify_num_parameters(7, 9, self.usage)

    # self.param packet [Packet] The packet the item should be added to
    def create_limits(self, packet, item, warnings):
        limits_set = self._get_limits_set()
        self._initialize_limits_values(packet, item)
        self._ensure_consistency_with_default(packet, item, warnings)

        item.limits.values[limits_set] = self._get_values()
        item.limits.enabled = self._get_enabled()
        item.limits.persistence_setting = self._get_persistence()
        item.limits.persistence_count = 0

        packet.update_limits_items_cache(item)
        return limits_set

    def _initialize_limits_values(self, packet, item):
        limits_set = self._get_limits_set()
        # Values must be initialized with a 'DEFAULT' key
        if item.limits.values is None:
            if limits_set == "DEFAULT":
                item.limits.values = {"DEFAULT": []}
            else:
                raise self.parser.error(
                    f"DEFAULT limits set must be defined for {packet.target_name} {packet.packet_name} {item.name} before setting limits set {limits_set}"
                )

    def _ensure_consistency_with_default(self, packet, item, warnings):
        # Nothing to do if we're already 'DEFAULT':
        if self._get_limits_set() == "DEFAULT":
            return

        msg = f"TELEMETRY Item {packet.target_name} {packet.packet_name} {item.name} {self._get_limits_set()} limits _TYPE_ setting conflict with DEFAULT"
        # XOR our setting with the items current setting
        # If it returns True then we have a mismatch and log the error
        if self._get_enabled() ^ item.limits.enabled:
            warnings.append(msg.replace("_TYPE_", "enable"))
            Logger.warn(warnings[-1])
        if item.limits.persistence_setting != self._get_persistence():
            warnings.append(msg.replace("_TYPE_", "persistence"))
            Logger.warn(warnings[-1])

    def _get_limits_set(self):
        return self.parser.parameters[0].upper()

    def _get_persistence(self):
        try:
            return int(self.parser.parameters[1])
        except ValueError as error:
            raise self.parser.error("Persistence must be an integer.", self.usage) from error

    def _get_enabled(self):
        enabled = self.parser.parameters[2].upper()
        if enabled != "ENABLED" and enabled != "DISABLED":
            raise self.parser.error("Initial LIMITS state must be ENABLED or DISABLED.", self.usage)
        if enabled == "ENABLED":
            return True
        else:
            return False

    def _get_values(self):
        values = self._get_red_yellow_values()
        return values + self._get_green_values(values[1], values[2])

    def _get_red_yellow_values(self):
        params = self.parser.parameters
        err = None
        try:
            err = "red low"
            red_low = float(params[3])
            err = "yellow low"
            yellow_low = float(params[4])
            err = "yellow high"
            yellow_high = float(params[5])
            err = "red high"
            red_high = float(params[6])
        except ValueError as error:
            raise self.parser.error(
                f"Invalid {err} limit value. Limits can be integers or floats.",
                self.usage,
            ) from error

        # Verify valid limits are specified
        if (red_low > yellow_low) or (yellow_low >= yellow_high) or (yellow_high > red_high):
            raise self.parser.error(
                "Invalid limits specified. Ensure yellow limits are within red limits.",
                self.usage,
            )

        return [red_low, yellow_low, yellow_high, red_high]

    def _get_green_values(self, yellow_low, yellow_high):
        params = self.parser.parameters
        # Since our initial parameter check verifies between 7 and 9 we do a
        # special check for 8 parameters which is an error
        if len(params) == 8:
            raise self.parser.error("Must give both a green low and green high value.", self.usage)
        if not len(params) == 9:
            return []

        try:
            err = "green low"
            green_low = float(params[7])
            err = "green high"
            green_high = float(params[8])
        except ValueError as error:
            raise self.parser.error(
                f"Invalid {err} limit value. Limits can be integers or floats.",
                self.usage,
            ) from error

        if (yellow_low > green_low) or (green_low >= green_high) or (green_high > yellow_high):
            raise self.parser.error(
                "Invalid limits specified. Ensure green limits are within yellow limits.",
                self.usage,
            )

        return [green_low, green_high]



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.


from openc3.conversions.conversion import Conversion


# Segmented polynomial conversions consist of polynomial conversions that are
# applied for a range of values.
class SegmentedPolynomialConversion(Conversion):
    # A polynomial conversion segment which applies the conversion from the
    # lower bound (inclusive) until another segment's lower bound is
    # encountered.
    class Segment:
        # Creates a polynomial conversion segment. Multiple Segments are used to
        # implement a {SegmentedPolynomialConversion}.
        #
        # @param lower_bound [Integer] The value at which point this polynomial conversion
        #   should apply. All values >= to this value will be converted using the
        #   given coefficients.
        # @param coeffs [Array<Integer>] The polynomial coefficients
        def __init__(self, lower_bound, coeffs):
            self.lower_bound = lower_bound
            self.coeffs = coeffs

        # Implement the comparison operator to compared based on the lower_bound
        # but sort in reverse order so the segment with the largest lower_bound
        # comes first. This makes the calculation code in call easier.
        #
        # @param other_segment [Segment] The segment to compare
        # @return [Integer] 1 if self.lower_bound > other_segment.lower_bound, 0
        #   if they are equal, -1 if self.lower_bound < other_segment.lower_bound
        def __lt__(self, other_segment):
            return other_segment.lower_bound < self.lower_bound

        # Implement equality operator primarily for ease of testing
        #
        # @param segment [Segment] Other segment
        def __eq__(self, other_segment):
            return self.lower_bound == other_segment.lower_bound and self.coeffs == other_segment.coeffs

        # Perform the polynomial conversion
        #
        # @param value [Numeric] The value to convert
        # @return [Float] The converted value
        def calculate(self, value):
            converted = 0.0
            for index in range(0, len(self.coeffs)):
                converted += float(self.coeffs[index]) * (value**index)
            return converted

    # Initialize the converted_type to :FLOAT and converted_bit_size to 64.
    #
    # @param segments [Array] Array of segments typically generated by as_json
    #   Format similar to the following: [[15, [3, 2]], [10, [1, 2]]]
    #   Where each entry is an array with the first value as the lower_bound
    #   and the other entry is an array of the coefficients for that segment.
    def __init__(self, segments=[]):
        super().__init__()
        self.segments = []
        self.params = []
        for lower_bound, coeffs in segments:
            self.add_segment(lower_bound, *coeffs)
        self.converted_type = "FLOAT"
        self.converted_bit_size = 64

    # Add a segment to the segmented polynomial. The lower bound is inclusive, but
    # is ignored for the segment with the lowest lower_bound.
    #
    # @param lower_bound [Integer] The value at which point this polynomial conversion
    #   should apply. All values >= to this value will be converted using the
    #   given coefficients.
    # @param coeffs [Array<Integer>] The polynomial coefficients
    def add_segment(self, lower_bound, *coeffs):
        self.params.append([lower_bound, coeffs])
        self.segments.append(SegmentedPolynomialConversion.Segment(lower_bound, coeffs))
        self.segments.sort()

    # @param (see Conversion#call)
    # @return [Float] The value with the polynomial applied
    def call(self, value, packet, buffer):
        # Try to find correct segment
        for segment in self.segments:
            if value >= segment.lower_bound:
                return segment.calculate(value)

        # Default to using segment with smallest lower_bound
        segment = self.segments[-1]
        if segment:
            return self.segments[-1].calculate(value)
        else:
            return None

    # @return [String] The name of the class followed by a description of all
    #   the polynomial segments.
    def __str__(self):
        result = ""
        count = 0
        for segment in self.segments:
            if count > 0:
                result += "\n"
            result += f"Lower Bound= {segment.lower_bound} Polynomial= "
            for index in range(0, len(segment.coeffs)):
                if index == 0:
                    result += f"{segment.coeffs[index]}"
                elif index == 1:
                    result += f" + {segment.coeffs[index]}x"
                else:
                    result += f" + {segment.coeffs[index]}x^{index}"
            count += 1
        return result

    # @param (see Conversion#to_config)
    # @return [String] Config fragment for this conversion
    def to_config(self, read_or_write):
        config = ""
        for segment in self.segments:
            config += f"    SEG_POLY_{read_or_write}_CONVERSION {segment.lower_bound} {' '.join(segment.coeffs)}\n"
        return config



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import os
import sys
from datetime import datetime, timezone


def quote_if_necessary(string: str):
    if " " in string:
        return f'"{string}"'
    else:
        return string


def simple_formatted(string: str):
    return "".join(format(x, "02x") for x in string).upper()


# The printable range of ASCII characters
PRINTABLE_RANGE = range(32, 127)


def formatted(
    input="",
    word_size=1,
    words_per_line=16,
    word_separator=" ",
    indent=0,
    show_address=True,
    address_separator=": ",
    show_ascii=True,
    ascii_separator="  ",
    unprintable_character=" ",
    line_separator="\n",
):
    string = ""
    byte_offset = 0
    bytes_per_line = word_size * words_per_line
    indent_string = " " * indent
    ascii_line = ""

    for byte in input:
        if byte_offset % bytes_per_line == 0:
            # Create the indentation at the try:ning of each line
            string += indent_string

            # Add the address if requested:
            if show_address:
                string += "%08X%s" % (byte_offset, address_separator)

        # Add the byte
        string += "%02X" % byte

        # Create the ASCII representation if requested:
        if show_ascii:
            if byte in PRINTABLE_RANGE:
                ascii_line += chr(byte)
            else:
                ascii_line += unprintable_character

        # Move to next byte
        byte_offset += 1

        # If we're at the end of the line we output the ascii if requested:
        if byte_offset % bytes_per_line == 0:
            if show_ascii:
                string += f"{ascii_separator}{ascii_line}"
                ascii_line = ""
                string += line_separator

        # If we're at a word junction then output the word_separator
        elif (byte_offset % word_size == 0) and byte_offset != len(input):
            string += word_separator

    # We're done printing all the bytes. Now check to see if we ended in the:
    # middle of a line. If so we have to print out the final ASCII if
    # requested.
    if byte_offset % bytes_per_line != 0:
        if show_ascii:
            num_word_separators = int(((byte_offset % bytes_per_line) - 1) / word_size)
            existing_length = (num_word_separators * len(word_separator)) + ((byte_offset % bytes_per_line) * 2)
            full_line_length = (bytes_per_line * 2) + ((words_per_line - 1) * len(word_separator))
            filler = " " * (full_line_length - existing_length)
            ascii_filler = " " * (bytes_per_line - len(ascii_line))
            string += f"{filler}{ascii_separator}{ascii_line}{ascii_filler}"
            ascii_line = ""
        string += line_separator
    return string


# Builds a String for use in creating a file. The time is formatted as
# YYYY_MM_DD_HH_MM_SS. The tags and joined with an underscore and appended to
# the date before appending the extension.
#
# For example:
#   File.build_timestamped_filename(['test','only'], '.bin', Time.now.sys)
#   # result is YYYY_MM_DD_HH_MM_SS_test_only.bin
#
# @param tags [Array<String>] An array of strings to be joined by underscores
#   after the date. Pass nil or an empty array to use no tags.
# @param extension [String] The filename extension
# @param time [Time] The time to format into the filename
# @return [String] The filename string containing the timestamp, tags, and
#   extension
def build_timestamped_filename(tags=None, extension=".txt", time=None):
    if time is None:
        time = datetime.now(timezone.utc)
    timestamp = time.strftime("%Y_%m_%d_%H_%M_%S")
    if not tags:
        tags = []
    tags = [str(t) for t in tags if t is not None]
    if len(tags) > 0:
        combined_tags = "_".join(tags)
        filename = timestamp + "_" + combined_tags + extension
    else:
        filename = timestamp + extension
    return filename


# Converts a String representing a class (i.e. "MyGreatClass") to a Python
# filename which implements the class (i.e. "my_great_class.py").
#
# self.param include_extension [Boolean] Whether to add '.py' extension
# self.return [String] Filename which implements the class name
def class_name_to_filename(string, include_extension=False):
    filename = ""
    for index in range(0, len(string)):
        if index != 0 and string[index] == string[index].upper():
            filename += "_"
        filename += string[index].lower()
    if include_extension:
        filename += ".py"
    return filename


# Converts a String representing a filename (i.e. "path/to/my_great_class.py")
# to a Python class name (i.e. "MyGreatClass").
#
# self.return [String] Class name associated with the filename
def filename_to_class_name(filename):
    return _snake_to_camel(os.path.basename(filename).split(".")[0])


# Converts a String representing a filename (i.e. "path/to/my_great_class.py")
# to a Python module name (i.e. "path.to.my_great_class").
#
# self.return [String] Module associated with the filename
def filename_to_module(filename):
    return filename.replace(".py", "").replace("/", ".")


def to_class(module, classname):
    if sys.modules.get(module):
        return getattr(sys.modules[module], classname)
    return None


def _snake_to_camel(snake):
    return "".join(x.capitalize() for x in snake.lower().split("_"))



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.script.server_proxy import ApiServerProxy, ScriptServerProxy
from openc3.utilities.authentication import OpenC3KeycloakAuthentication
from openc3.utilities.extract import convert_to_value
from openc3.environment import OPENC3_KEYCLOAK_URL
import threading

API_SERVER = ApiServerProxy()
SCRIPT_RUNNER_API_SERVER = ScriptServerProxy()
RUNNING_SCRIPT = None
DISCONNECT = False
OPENC3_IN_CLUSTER = False
if "openc3-cosmos-cmd-tlm-api" in API_SERVER.generate_url():
    OPENC3_IN_CLUSTER = True


def shutdown_script():
    global API_SERVER
    API_SERVER.shutdown()
    global SCRIPT_RUNNER_API_SERVER
    SCRIPT_RUNNER_API_SERVER.shutdown()


def prompt_for_hazardous(target_name, cmd_name, hazardous_description):
    """ """
    message_list = [f"Warning: Command {target_name} {cmd_name} is Hazardous. "]
    if hazardous_description:
        message_list.append(hazardous_description)
    message_list.append("Send? (y/N): ")
    answer = input("\n".join(message_list))
    try:
        return answer.lower()[0] == "y"
    except IndexError:
        return False


def _file_dialog(title, message, filter=None):
    answer = ""
    while len(answer) == 0:
        print(f"{title}\n{message}\n<Type file name>:")
        answer = input()
    return answer


###########################################################################
# START PUBLIC API
###########################################################################


def disconnect_script():
    global DISCONNECT
    DISCONNECT = True


def ask_string(question, blank_or_default=False, password=False):
    answer = ""
    default = None
    if blank_or_default is not True and blank_or_default is not False:
        question += f" (default = {blank_or_default})"
        default = str(blank_or_default)
        allow_blank = True
    else:
        allow_blank = blank_or_default
    while len(answer) == 0:
        print(question + " ")
        answer = input()
        if allow_blank:
            break
    if len(answer) == 0 and default:
        answer = default
    return answer


def ask(question, blank_or_default=False, password=False):
    string = ask_string(question, blank_or_default, password)
    value = convert_to_value(string)
    return value


def message_box(string, *buttons, **options):
    print(f"{string} ({', '.join(buttons)}): ")
    if "details" in options:
        print(f"Details: {options['details']}\n")
    return input()


def vertical_message_box(string, *buttons, **options):
    return message_box(string, *buttons, **options)


def combo_box(string, *buttons, **options):
    return message_box(string, *buttons, **options)


def open_file_dialog(title, message="Open File", filter=None):
    _file_dialog(title, message, filter)


def open_files_dialog(title, message="Open File", filter=None):
    _file_dialog(title, message, filter)


def prompt(
    string,
    text_color=None,
    background_color=None,
    font_size=None,
    font_family=None,
    details=None,
):
    print(f"{string}: ")
    if details:
        print(f"Details: {details}\n")
    return input()


def step_mode():
    # running_script.py implements the real functionality
    pass


def run_mode():
    # running_script.py implements the real functionality
    pass

from .api_shared import *
from .commands import *
from .cosmos_calendar import *
from .critical_cmd import *
from .exceptions import *
from .limits import *
from .metadata import *
from .packages import *
from .plugins import *
from .screen import *
from .script_runner import *
from .storage import *
from .tables import *
from .telemetry import *
from openc3.api import WHITELIST

# Note: Enterprise Only - Use this for first time setup of an offline access token
# so that users can run scripts.  Not necessary if accessing APIs via the web
# frontend as it handles it automatically.
#
# Example:
# initialize_offline_access()
# script_run("INST/procedures/collect.rb")
#
def initialize_offline_access():
    auth = OpenC3KeycloakAuthentication(OPENC3_KEYCLOAK_URL)
    auth.token(include_bearer=True, openid_scope='openid%20offline_access')
    set_offline_access(auth.refresh_token)

###########################################################################
# END PUBLIC API
###########################################################################


def prompt_for_critical_cmd(uuid, _username, _target_name, _cmd_name, _cmd_params, cmd_string):
    print("Waiting for critical command approval:")
    print(f"  {cmd_string}")
    print(f"  UUID: {uuid}")
    while True:
        status = critical_cmd_status(uuid)
        if status == "APPROVED":
            return
        elif status == "REJECTED":
            raise RuntimeError("Critical command rejected")
        threading.sleep(0.1)


# Define all the WHITELIST methods
current_functions = dir()
for func in WHITELIST:
    if func not in current_functions:
        code = f"def {func}(*args, **kwargs):\n    return getattr(API_SERVER, '{func}')(*args, **kwargs)"
        function = compile(code, "<string>", "exec")
        exec(function, globals())



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from .suite import Suite, Group, ScriptStatus
from openc3.tools.test_runner.test import TestSuite, Test
from .suite_results import SuiteResults
from openc3.script.exceptions import StopScript
import inspect

class UnassignedSuite(Suite):
    """Placeholder for all Groups discovered without assigned Suites"""
    pass


class SuiteRunner:
    suites = []
    settings = {}
    suite_results = None

    @classmethod
    def execute(cls, result_string, suite_class, group_class=None, script=None):
        SuiteRunner.suite_results = SuiteResults()
        for suite in SuiteRunner.suites:
            if suite.__class__ == suite_class:
                SuiteRunner.suite_results.start(
                    result_string,
                    suite_class,
                    group_class,
                    script,
                    SuiteRunner.settings,
                )
                while True:
                    yield (suite)
                    if not SuiteRunner.settings["Loop"] or (
                        ScriptStatus.instance().fail_count > 0 and SuiteRunner.settings["Break Loop On Error"]
                    ):
                        break
                break

    @classmethod
    def start(cls, suite_class, group_class=None, script=None):
        result = []
        for suite in cls.execute("", suite_class, group_class, script):
            if script:
                result = suite.run_script(group_class, script)
                SuiteRunner.suite_results.process_result(result)
                if (result.exceptions and SuiteRunner.settings["Abort After Error"]) or result.stopped:
                    raise StopScript
            elif group_class:
                for result in suite.run_group(group_class):
                    SuiteRunner.suite_results.process_result(result)
                    if result.stopped:
                        raise StopScript
            else:
                for result in suite.run():
                    SuiteRunner.suite_results.process_result(result)
                    if result.stopped:
                        raise StopScript

    @classmethod
    def setup(cls, suite_class, group_class=None):
        for suite in cls.execute("Manual Setup", suite_class, group_class):
            if group_class:
                result = suite.run_group_setup(group_class)
            else:
                result = suite.run_setup()

            if result:
                SuiteRunner.suite_results.process_result(result)
                if result.stopped:
                    raise StopScript

    @classmethod
    def teardown(cls, suite_class, group_class=None):
        for suite in cls.execute("Manual Teardown", suite_class, group_class):
            if group_class:
                result = suite.run_group_teardown(group_class)
            else:
                result = suite.run_teardown()

            if result:
                SuiteRunner.suite_results.process_result(result)
                if result.stopped:
                    raise StopScript

    # Build list of Suites and Groups
    @classmethod
    def build_suites(cls, from_module=None, from_globals=None):
        SuiteRunner.suites = []
        suites = {}
        groups = []

        if from_module:
            for attr_name in dir(from_module):
                object = getattr(from_module, attr_name)
                if not inspect.isclass(object):
                    continue

                # If we inherit from Suite
                if issubclass(object, Suite) and object != Suite and object != TestSuite:
                    SuiteRunner.suites.insert(0, object())

                # If we inherit from Group
                if issubclass(object, Group) and object != Group and object != Test:
                    groups.append(object)

        if from_globals:
            for object in from_globals.values():
                if not inspect.isclass(object):
                    continue

                # If we inherit from Suite
                if issubclass(object, Suite) and object != Suite and object != TestSuite:
                    SuiteRunner.suites.insert(0, object())

                # If we inherit from Group
                if issubclass(object, Group) and object != Group and object != Test:
                    groups.append(object)

        # Raise error if no suites or groups
        if len(SuiteRunner.suites) == 0 or len(groups) == 0:
            return "No Suite or no Group classes found"

        # Remove assigned Groups from the array of groups
        for suite in SuiteRunner.suites:
            if suite.__class__ == UnassignedSuite:
                continue
            groups_to_delete = []
            for group in groups:
                if group in suite.scripts():
                    groups_to_delete.append(group)
            for group in groups_to_delete:
                groups.remove(group)

        if len(groups) == 0:
            # If there are no unassigned group we simply remove the UnassignedSuite
            SuiteRunner.suites = [suite for suite in SuiteRunner.suites if suite.__class__ != UnassignedSuite]
        else:
            # unassigned groups should be added to the UnassignedSuite
            unassigned_suite = UnassignedSuite()
            for group in groups:
                unassigned_suite.add_group(group)

        for suite in SuiteRunner.suites:
            cur_suite = {"setup": False, "teardown": False, "groups": {}}
            if "setup" in dir(suite):
                cur_suite["setup"] = True
            if "teardown" in dir(suite):
                cur_suite["teardown"] = True

            for type, group_class, script in suite.plans():
                match type:
                    case "GROUP":
                        if not cur_suite["groups"].get(group_class.__name__):
                            cur_suite["groups"][group_class.__name__] = {
                                "setup": False,
                                "teardown": False,
                                "scripts": [],
                            }
                        cur_suite["groups"][group_class.__name__]["scripts"].extend(group_class.scripts())
                        # Make uniq!
                        temp = set(cur_suite["groups"][group_class.__name__]["scripts"])
                        cur_suite["groups"][group_class.__name__]["scripts"] = list(temp)
                        cur_suite["groups"][group_class.__name__]["scripts"].sort()
                        if "setup" in dir(group_class):
                            cur_suite["groups"][group_class.__name__]["setup"] = True
                        if "teardown" in dir(group_class):
                            cur_suite["groups"][group_class.__name__]["teardown"] = True
                    case "SCRIPT":
                        if not cur_suite["groups"].get(group_class.__name__):
                            cur_suite["groups"][group_class.__name__] = {
                                "setup": False,
                                "teardown": False,
                                "scripts": [],
                            }
                        # Explicitly check for this method and raise an error if it does not exist
                        if script in dir(group_class):
                            cur_suite["groups"][group_class.__name__]["scripts"].append(script)
                            # Make uniq!
                            temp = set(cur_suite["groups"][group_class.__name__]["scripts"])
                            cur_suite["groups"][group_class.__name__]["scripts"] = list(temp)
                            cur_suite["groups"][group_class.__name__]["scripts"].sort()
                        else:
                            raise Exception(f"{group_class} does not have a {script} method defined.")

                        if "setup" in dir(group_class):
                            cur_suite["groups"][group_class.__name__]["setup"] = True
                        if "teardown" in dir(group_class):
                            cur_suite["groups"][group_class.__name__]["teardown"] = True
                    case "GROUP_SETUP":
                        if not cur_suite["groups"].get(group_class.__name__):
                            cur_suite["groups"][group_class.__name__] = {
                                "setup": False,
                                "teardown": False,
                                "scripts": [],
                            }
                        # Explicitly check for the setup method and raise an error if it does not exist
                        if "setup" in dir(group_class):
                            cur_suite["groups"][group_class.__name__]["setup"] = True
                        else:
                            raise Exception(f"{group_class} does not have a setup method defined.")

                    case "GROUP_TEARDOWN":
                        if not cur_suite["groups"].get(group_class.__name__):
                            cur_suite["groups"][group_class.__name__] = {
                                "setup": False,
                                "teardown": False,
                                "scripts": [],
                            }
                        # Explicitly check for the teardown method and raise an error if it does not exist
                        if "teardown" in dir(group_class):
                            cur_suite["groups"][group_class.__name__]["teardown"] = True
                        else:
                            raise Exception(f"{group_class} does not have a teardown method defined.")

            if not suite.name == "CustomSuite":
                suites[suite.name()] = cur_suite

        return suites



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import re
import time
from datetime import datetime, timezone
from openc3.utilities.extract import remove_quotes
import traceback


class SuiteResults:
    metadata = None
    context = None

    def __init__(self):
        self._report = None
        self.context = None
        self.start_time = None
        self.stop_time = None
        self.results = None
        self.settings = None
        self.metadata = None

    def start(
        self,
        test_type,
        test_suite_class,
        test_class=None,
        test_case=None,
        settings=None,
    ):
        self.results = []
        self.start_time = time.time()
        self.settings = settings
        self._report = []

        if test_case:
            # Executing a script
            self.context = f"{test_suite_class.__name__}:{test_class.__name__}:{test_case} {test_type}"
        elif test_class:
            # Executing a group
            self.context = f"{test_suite_class.__name__}:{test_class.__name__} {test_type}"
        else:
            # Executing a suite
            self.context = f"{test_suite_class.__name__} {test_type}"
        self.header()

    # process_result can handle an array of OpenC3TestResult objects
    # or a single OpenC3TestResult object
    def process_result(self, results):
        # If we were passed an array we concat it to the results global
        if isinstance(results, list):
            self.results.append(results)
        # A single result is appended and then turned into an array
        else:
            self.results.append(results)
            results = [results]

        # Process all the results (may be just one)
        for result in results:
            self.puts(f"{result.group}:{result.script}:{result.result}")
            if result.message:
                for line in result.message.split("\n"):
                    if re.search(r"\x00-\x08\x0B-\x0C\x0E-\x1F\x7F-\xFF", line):
                        line = line.rstrip("\n")
                        line = remove_quotes(repr(line))
                    self._report.append("  " + line.strip())

            if result.exceptions:
                self._report.append("  Exceptions:")
                for _, error in enumerate(result.exceptions):
                    self._report.append("".join(traceback.format_exception(error)))

    def complete(self):
        self.stop_time = time.time()
        self.footer()

    def report(self):
        return "\n".join(self._report)

    def header(self):
        self._report.append("--- Script Report ---")
        if self.settings:
            self._report.append("")
            self._report.append("Settings:")
            for setting_name, setting_value in self.settings.items():
                self._report.append(f"{setting_name} = {setting_value}")

        self._report.append("")
        self._report.append("Results:")
        self.puts(f"Executing {self.context}")

    def footer(self):
        self.puts(f"Completed {self.context}")

        self._report.append("")
        self._report.append("--- Test Summary ---")
        self._report.append("")

        pass_count = 0
        skip_count = 0
        fail_count = 0
        stopped = False
        for result in self.results:
            if result.result == "PASS":
                pass_count += 1
            elif result.result == "SKIP":
                skip_count += 1
            elif result.result == "FAIL":
                fail_count += 1
            if result.stopped:
                stopped = True

        run_time = self.stop_time - self.start_time
        self._report.append(f"Run Time: {run_time}")
        self._report.append(f"Total Tests: {len(self.results)}")
        self._report.append(f"Pass: {pass_count}")
        self._report.append(f"Skip: {skip_count}")
        self._report.append(f"Fail: {fail_count}")
        self._report.append("")
        if stopped:
            self._report.append("*** Test was stopped prematurely ***")
            self._report.append("")

    def write(self, string):
        # Can't use isoformat because it appends "+00:00" instead of "Z"
        self._report.append(datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%fZ") + ": " + string)

    # Define a few aliases
    puts = write
    print = write



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.
#
# A portion of this file was funded by Blue Origin Enterprises, L.P.
# See https://github.com/OpenC3/cosmos/pull/1953


class Telemetry:
    """Telemetry uses PacketConfig to parse the command and telemetry
    configuration files. It contains all the knowledge of which telemetry packets
    exist in the system and how to access them. This class is the API layer
    which other classes use to access telemetry.

    This should not be confused with the Api module which implements the JSON
    API that is used by tools when accessing the Server. The Api module always
    provides Ruby primitives where the Telemetry class can return actual
    Packet or PacketItem objects. While there are some overlapping methods between
    the two, these are separate interfaces into the system."""

    LATEST_PACKET_NAME = "LATEST"

    # @param config [PacketConfig] Packet configuration to use to access the
    #  telemetry
    def __init__(self, config, system):
        self.system = system
        self.config = config

    # (see PacketConfig#warnings)
    def warnings(self):
        return self.config.warnings

    # @return [Array<String>] The command target names (excluding UNKNOWN)
    def target_names(self):
        result = list(self.config.telemetry.keys())
        result.remove("UNKNOWN")
        result.sort()
        return result

    # @param target_name [String] The target name
    # @return [Hash<packet_name=>Packet>] Hash of the telemetry packets for the given
    #   target name keyed by the packet name
    def packets(self, target_name):
        upcase_target_name = target_name.upper()
        target_packets = self.config.telemetry.get(upcase_target_name, None)
        if not target_packets:
            raise RuntimeError(f"Telemetry target '{upcase_target_name}' does not exist")

        return target_packets

    # @param target_name [String] The target name
    # @param packet_name [String] The packet name. Must be a defined packet name
    #   and not 'LATEST'.
    # @return [Packet] The telemetry packet for the given target and packet name
    def packet(self, target_name, packet_name):
        target_packets = self.packets(target_name)
        upcase_packet_name = packet_name.upper()
        packet = target_packets.get(upcase_packet_name, None)
        if not packet:
            upcase_target_name = target_name.upper()
            raise RuntimeError(f"Telemetry packet '{upcase_target_name} {upcase_packet_name}' does not exist")
        return packet

    # Identifies an unknown buffer of data as a defined packet and sets the
    # packet's data to the given buffer. Identifying a packet uses the fields
    # marked as ID_ITEM to identify if the buffer passed represents the
    # packet defined. Incorrectly sized buffers are still processed but an
    # error is logged.
    #
    # Note: This affects all subsequent requests for the packet (for example
    # using packet)
    #
    # @param packet_data [String] The binary packet data buffer
    # @param target_names [Array<String>] List of target names to limit the search. The
    #   default value of nil means to search all known targets.
    # @return [Packet] The identified packet with its data set to the given
    #   packet_data buffer. Returns nil if no packet could be identified.
    def identify_and_set_buffer(self, packet_data, target_names=None):
        identified_packet = self.identify(packet_data, target_names)
        if identified_packet:
            identified_packet.buffer = packet_data
        return identified_packet

    # Finds a packet from the Current Value Table that matches the given data
    # and returns it.  Does not fill the packets buffer.  Use identify! to update the CVT.
    #
    # @param packet_data [String] The binary packet data buffer
    # @param target_names [Array<String>] List of target names to limit the search. The
    #   default value of nil means to search all known targets.
    # @return [Packet] The identified packet, Returns nil if no packet could be identified.
    def identify(self, packet_data, target_names=None):
        if target_names is None:
            target_names = self.target_names()

        for target_name in target_names:
            target_name = str(target_name).upper()

            target_packets = None
            try:
                target_packets = self.packets(target_name)
            except RuntimeError:
                # No telemetry for this target
                continue

            target = self.system.targets.get(target_name)
            if target and target.tlm_unique_id_mode:
                # Iterate through the packets and see if any represent the buffer
                for _, packet in target_packets.items():
                    if packet.identify(packet_data):
                        return packet
            else:
                # Do a hash lookup to quickly identify the packet
                if len(target_packets) > 0:
                    packet = next(iter(target_packets.values()))
                    key = packet.read_id_values(packet_data)
                    id_values = self.config.tlm_id_value_hash[target_name]
                    identified_packet = id_values.get(repr(key))
                    if identified_packet is None:
                        identified_packet = id_values.get("CATCHALL")
                    if identified_packet is not None:
                        return identified_packet

        return None

    def identify_and_define_packet(self, packet, target_names=None):
        if not packet.identified():
            identified_packet = self.identify(packet.buffer_no_copy(), target_names)
            if not identified_packet:
                return None

            identified_packet = identified_packet.clone()
            identified_packet.buffer = packet.buffer
            identified_packet.received_time = packet.received_time
            identified_packet.stored = packet.stored
            identified_packet.extra = packet.extra
            return identified_packet

        if not packet.defined():
            try:
                identified_packet = self.packet(packet.target_name, packet.packet_name)
            except RuntimeError:
                return None
            identified_packet = identified_packet.clone()
            identified_packet.buffer = packet.buffer
            identified_packet.received_time = packet.received_time
            identified_packet.stored = packet.stored
            identified_packet.extra = packet.extra
            return identified_packet

        return packet

    # Updates the specified packet with the given packet data. Raises an error
    # if the packet could not be found.
    #
    # Note: This affects all subsequent requests for the packet
    #
    # @param target_name (see #packet)
    # @param packet_name (see #packet)
    # @param packet_data (see #identify_tlm!)
    # @return [Packet] The packet with its data set to the given packet_data
    #   buffer.
    def update(self, target_name, packet_name, packet_data):
        identified_packet = self.packet(target_name, packet_name)
        identified_packet.buffer = packet_data
        return identified_packet

    # Assigns a limits change callback to all telemetry packets
    #
    # @param limits_change_callback
    def set_limits_change_callback(self, limits_change_callback):
        for _, packets in self.config.telemetry.items():
            for _, packet in packets.items():
                packet.limits_change_callback = limits_change_callback

    # Resets metadata on every packet in every target
    def reset(self):
        for _, packets in self.config.telemetry.items():
            for _, packet in packets.items():
                packet.reset()

    # @return [Hash{String=>Hash{String=>Packet}}] Hash of all the telemetry
    #   packets keyed by the target name. The value is another hash keyed by the
    #   packet name returning the packet.
    def all(self):
        return self.config.telemetry

    def dynamic_add_packet(self, packet, affect_ids=False):
        self.config.dynamic_add_packet(packet, "TELEMETRY", affect_ids=affect_ids)



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import socket
import sys
import json
from datetime import datetime, timezone
from threading import Lock
from openc3.environment import *
from openc3.utilities.store_queued import EphemeralStoreQueued

# Logger class, class attribute list
CLASS_ATTRS = [
    "instance",
    "instance_mutex",
    "my_instance",
    "scope",
    "__dict__",
    "DEBUG",
    "INFO",
    "WARN",
    "ERROR",
    "FATAL",
    "DEBUG_LEVEL",
    "INFO_LEVEL",
    "WARN_LEVEL",
    "ERROR_LEVEL",
    "FATAL_LEVEL",
    "LOG",
    "NOTIFICATION",
    "ALERT",
]


# Logger class, instance attribute list
INSTANCE_ATTRS = [
    "stdout",
    "level",
    "detail_string",
    "container_name",
    "microservice_name",
    "no_store",
]


class LoggerMeta(type):
    def __getattribute__(cls, func):
        if func in CLASS_ATTRS:
            return super().__getattribute__(func)

        if func in INSTANCE_ATTRS:
            return getattr(cls.instance(), func)

        # Handle dunder methods to support help() and other introspection
        if func.startswith("__") and func.endswith("__"):
            return super().__getattribute__(func)

        def method(*args, **kw_args):
            return getattr(cls.instance(), func)(*args, **kw_args)

        return method

    def __setattr__(cls, func, value):
        if func in INSTANCE_ATTRS:
            return setattr(cls.instance(), func, value)

        if func in CLASS_ATTRS:
            return super().__setattr__(func, value)

        raise AttributeError(f"Unknown attribute {func}")


# Supports different levels of logging and only writes if the level
# is exceeded.
class Logger(metaclass=LoggerMeta):
    instance_mutex = Lock()
    my_instance = None
    scope = OPENC3_SCOPE

    DEBUG = 0
    INFO = 1
    WARN = 2
    ERROR = 3
    FATAL = 4

    DEBUG_LEVEL = "DEBUG"
    INFO_LEVEL = "INFO"
    WARN_LEVEL = "WARN"
    ERROR_LEVEL = "ERROR"
    FATAL_LEVEL = "FATAL"

    LOG = "log"
    NOTIFICATION = "notification"
    ALERT = "alert"
    EPHEMERAL = "ephemeral"

    # @param level [Integer] The initial logging level
    def __init__(self, level=INFO):
        self.stdout = True
        self.level = level
        self.detail_string = None
        self.container_name = socket.gethostname()
        self.microservice_name = None
        if OPENC3_NO_STORE:
            self.no_store = True
        else:
            self.no_store = False

    # Get the singleton instance
    @classmethod
    def instance(cls, level=INFO):
        if cls.my_instance:
            return cls.my_instance

        with cls.instance_mutex:
            cls.my_instance = cls(level)
            return cls.my_instance

    # @param message [String] The message to print if the log level is at or
    #   below the method name log level.
    # @param block [Proc] Block to call which should return a string to append
    #   to the log message
    def debug(self, message=None, scope=None, user=None, type=LOG, url=None, other=None):
        scope = scope or self.scope
        if self.level <= self.DEBUG:
            self.log_message(
                self.DEBUG_LEVEL,
                message,
                scope=scope,
                user=user,
                type=type,
                url=url,
                other=other,
            )

    # (see #debug)
    def info(self, message=None, scope=None, user=None, type=LOG, url=None, other=None):
        scope = scope or self.scope
        if self.level <= self.INFO:
            self.log_message(
                self.INFO_LEVEL,
                message,
                scope=scope,
                user=user,
                type=type,
                url=url,
                other=other,
            )

    # (see #debug)
    def warn(self, message=None, scope=None, user=None, type=LOG, url=None, other=None):
        scope = scope or self.scope
        if self.level <= self.WARN:
            self.log_message(
                self.WARN_LEVEL,
                message,
                scope=scope,
                user=user,
                type=type,
                url=url,
                other=other,
            )

    # (see #debug)
    def error(self, message=None, scope=None, user=None, type=LOG, url=None, other=None):
        scope = scope or self.scope
        if self.level <= self.ERROR:
            self.log_message(
                self.ERROR_LEVEL,
                message,
                scope=scope,
                user=user,
                type=type,
                url=url,
                other=other,
            )

    # (see #debug)
    def fatal(self, message=None, scope=None, user=None, type=LOG, url=None, other=None):
        scope = scope or self.scope
        if self.level <= self.FATAL:
            self.log_message(
                self.FATAL_LEVEL,
                message,
                scope=scope,
                user=user,
                type=type,
                url=url,
                other=other,
            )

    def build_log_data(self, log_level, message, user=None, type=None, url=None, other=None):
        now_time = datetime.now(timezone.utc)
        data = {
            "time": int(now_time.timestamp() * 1000000000),
            # Can't use isoformat because it appends "+00:00" instead of "Z"
            "@timestamp": now_time.strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
            "level": log_level,
        }
        if self.microservice_name is not None:
            data["microservice_name"] = self.microservice_name
        if self.detail_string is not None:
            data["detail"] = self.detail_string
        # EE: If a user is passed, put its name. Don't include user data if no user was passed.
        if user is not None:
            data["user"] = user
        data["container_name"] = self.container_name
        if message is not None:
            data["message"] = message
        if type is not None:
            data["type"] = type
        if url is not None:
            data["url"] = url
        if other is not None:
            data = data | other
        return data

    def log_message(self, log_level, message, scope, user, type, url, other=None):
        with self.instance_mutex:
            data = self.build_log_data(log_level, message, user, type, url, other)
            if self.stdout:
                match log_level:
                    case "WARN" | "ERROR" | "FATAL":
                        if OPENC3_LOG_STDERR:
                            print(json.dumps(data), file=sys.stderr)
                            sys.stderr.flush()
                        else:
                            print(json.dumps(data), file=sys.stdout)
                            sys.stdout.flush()
                    case _:
                        print(json.dumps(data), file=sys.stdout)
                        sys.stdout.flush()
            if self.no_store is False:
                if scope is not None:
                    EphemeralStoreQueued.write_topic(f"{scope}__openc3_log_messages", data)
                else:
                    # The base openc3_log_messages doesn't have an associated logger
                    # so it must be limited to prevent unbounded stream growth
                    EphemeralStoreQueued.write_topic("NOSCOPE__openc3_log_messages", data)



# Copyright 2023 OpenC3, Inc
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.:

import json
from openc3.api import WHITELIST
from openc3.environment import OPENC3_SCOPE
from openc3.utilities.authorization import authorize
from openc3.models.tool_config_model import ToolConfigModel

WHITELIST.extend(["config_tool_names", "list_configs", "load_config", "save_config", "delete_config"])


def config_tool_names(scope=OPENC3_SCOPE):
    authorize(permission="system", scope=scope)
    return ToolConfigModel.config_tool_names(scope=scope)


def list_configs(tool, scope=OPENC3_SCOPE):
    authorize(permission="system", scope=scope)
    return ToolConfigModel.list_configs(tool, scope=scope)


def load_config(tool, name, scope=OPENC3_SCOPE):
    authorize(permission="system", scope=scope)
    return json.loads(ToolConfigModel.load_config(tool, name, scope=scope))


def save_config(tool, name, data, local_mode=True, scope=OPENC3_SCOPE):
    authorize(permission="system_set", scope=scope)
    ToolConfigModel.save_config(tool, name, data, local_mode, scope)


def delete_config(tool, name, local_mode=True, scope=OPENC3_SCOPE):
    authorize(permission="system_set", scope=scope)
    ToolConfigModel.delete_config(tool, name, local_mode, scope)



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.packets.packet_item import PacketItem

class TableItem(PacketItem):
    """
    Implements the attributes that are unique to a TableItem such as editable
    and hidden. All other functionality is inherited from PacketItem.
    """

    def __init__(
        self,
        name,
        bit_offset,
        bit_size,
        data_type,
        endianness,
        array_size=None,
        overflow="ERROR"
    ):
        """
        Constructor

        Args:
            name: The name of the item. Used by the accessor methods to grab data.
            bit_offset: The offset in bits where this item is found in the packet
            bit_size: The size in bits of this item
            data_type: The data type of this item. Must be one of :INT, :UINT, :FLOAT, or :STRING, :BLOCK
            endianness: Must be one of :BIG_ENDIAN or :LITTLE_ENDIAN
            array_size: Set to a value if this item is to be an array. Default is nil.
            overflow: How to handle value overflows. Must be one of :ERROR, :ERROR_ALLOW_HEX, :TRUNCATE, :SATURATE
        """
        super().__init__(
            name,
            bit_offset,
            bit_size,
            data_type,
            endianness,
            array_size,
            overflow,
        )
        self.display_type = None
        self._editable = True
        self._hidden = False

    @property
    def editable(self):
        """Whether this item is editable"""
        return self._editable

    @editable.setter
    def editable(self, editable):
        """
        Set whether this item can be edited

        Args:
            editable: Boolean indicating if item is editable

        Raises:
            TypeError if editable is not a boolean
        """
        if not isinstance(editable, bool):
            raise TypeError(f"{self.name}: editable must be a boolean but is a {type(editable).__name__}")
        self._editable = editable

    @property
    def hidden(self):
        """Whether this item is hidden (not displayed)"""
        return self._hidden

    @hidden.setter
    def hidden(self, hidden):
        """
        Set whether this item should be hidden

        Args:
            hidden: Boolean indicating if item is hidden

        Raises:
            TypeError if hidden is not a boolean
        """
        if not isinstance(hidden, bool):
            raise TypeError(f"{self.name}: hidden must be a boolean but is a {type(hidden).__name__}")
        self._hidden = hidden

    def clone(self):
        """
        Make a light weight clone of this item

        Returns:
            Cloned TableItem
        """
        item = super().clone()
        item.editable = self.editable
        item.hidden = self.hidden
        return item

    def as_json(self, *args, **kwargs):
        """
        Create a hash of this item's attributes

        Returns:
            Dict representing the item
        """
        hash_val = super().as_json(*args, **kwargs)
        hash_val['editable'] = self.editable
        hash_val['hidden'] = self.hidden
        return hash_val


# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt

# Modified by OpenC3, Inc.
# All changes Copyright 2023, OpenC3, Inc.
# All Rights Reserved
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import time
from typing import Optional
import openc3.script
from openc3.script.api_shared import openc3_script_sleep
from openc3.environment import OPENC3_SCOPE


def get_packets(
    id: str, block: Optional[float] = None, block_delay: float = 0.1, count: int = 1000, scope: str = OPENC3_SCOPE
):
    """Get packets based on ID returned from subscribe_packet.

    Args:
        id (str) ID returned from subscribe_packets or last call to get_packets
        block (float) Time in seconds to wait for packets to be received
        block_delay (float) Time in seconds to sleep between polls
        count (int) Maximum number of packets to return from EACH packet stream
        scope (str) Optional, defaults to env.OPENC3_SCOPE

    Return:
        [Array<String, Array<Hash>] Array of the ID and array of all packets found
    """
    start_time = time.time()
    while True:
        id, packets = getattr(openc3.script.API_SERVER, "get_packets")(id, count=count, scope=scope)
        if block and time.time() < (start_time + block) and not packets:
            openc3_script_sleep(block_delay)
        else:
            break
    return id, packets


# inject_tlm, set_tlm, override_tlm, and normalize_tlm are implemented here simply to add a print
# these methods modify the telemetry so the user should be notified in the Script Runner log messages


def inject_tlm(
    target_name: str, packet_name: str, item_hash: dict = None, type: str = "CONVERTED", scope: str = OPENC3_SCOPE
):
    print(f'inject_tlm("{target_name}", "{packet_name}", {item_hash}, type="{type}")')
    getattr(openc3.script.API_SERVER, "inject_tlm")(target_name, packet_name, item_hash, type=type, scope=scope)


def set_tlm(*args, type: str = "CONVERTED", scope: str = OPENC3_SCOPE):
    if len(args) == 1:
        print(f'set_tlm("{args[0]}", type="{type}")')
    else:
        if isinstance(args[3], str):
            value = f'"{args[3]}"'
        else:
            value = args[3]
        print(f'set_tlm("{args[0]}", "{args[1]}", "{args[2]}", {value}, type="{type}")')
    getattr(openc3.script.API_SERVER, "set_tlm")(*args, type=type, scope=scope)


def override_tlm(*args, type: str = "ALL", scope: str = OPENC3_SCOPE):
    if len(args) == 1:
        print(f'override_tlm("{args[0]}", type="{type}")')
    else:
        if isinstance(args[3], str):
            value = f'"{args[3]}"'
        else:
            value = args[3]
        print(f'override_tlm("{args[0]}", "{args[1]}", "{args[2]}", {value}, type="{type}")')
    getattr(openc3.script.API_SERVER, "override_tlm")(*args, type=type, scope=scope)


def normalize_tlm(*args, type: str = "ALL", scope: str = OPENC3_SCOPE):
    if len(args) == 1:
        print(f'normalize_tlm("{args[0]}", type="{type}")')
    else:
        if isinstance(args[3], str):
            value = f'"{args[3]}"'
        else:
            value = args[3]
        print(f'normalize_tlm("{args[0]}", "{args[1]}", "{args[2]}", {value}, type="{type}")')
    getattr(openc3.script.API_SERVER, "normalize_tlm")(*args, type=type, scope=scope)



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import copy
from openc3.packets.packet import Packet


class Processor:
    # Create a new Processor
    # self.param value_type [Symbol or String] the value type to process
    def __init__(self, value_type="CONVERTED"):
        self.name = self.__class__.__name__.upper()
        value_type = value_type.upper()
        self.value_type = value_type
        if self.value_type not in Packet.VALUE_TYPES:
            raise ValueError(f"value_type must be RAW, CONVERTED, FORMATTED, or WITH_UNITS. Is {self.value_type}")

        self.results = {}

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, name):
        self.__name = name.upper()

    # Perform processing on the packet.
    #
    # self.param packet [Packet] The packet which contains the value. This can
    #   be useful to reach into the packet and use other values in the
    #   conversion.
    # self.param buffer [String] The packet buffer
    # self.return The processed result
    def call(self, packet, buffer):
        raise RuntimeError("call method must be defined by subclass")

    def __str__(self):
        return self.__class__.__name__

    # Reset any state
    def reset(self):
        pass
        # By default do nothing

    # Make a light weight clone of this processor. This only creates a new hash of results
    #
    # self.return [Processor] A copy of the processor with a new hash of results
    def clone(self):
        return copy.deepcopy(self)

    # Convert to configuration file string
    def to_config(self):
        return f"  PROCESSOR {self.name} {self.__class__.__name__} {self.value_type}\n"

    def as_json(self):
        return {
            "name": self.name,
            "class": self.__class__.__name__,
            "params": [self.value_type],
        }



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
from requests import Session
from threading import Lock
from typing import Optional
from openc3.environment import *
from openc3.utilities.authentication import (
    OpenC3Authentication,
    OpenC3KeycloakAuthentication,
)


class JsonApiError(Exception):
    pass


class JsonApiObject:
    """Used to forward all method calls to the remote server object. Before using
    this class ensure the remote service has been started in the server class:

        json = JsonApiObject('127.0.0.1', 7777, self)

    Now the JsonApiObject can be used to call server methods directly:

        server = JsonApiObject('http://openc3-cosmos-cmd-tlm-api:2901', 1.0)
        server.cmd(*args)

    """

    USER_AGENT = "OpenC3 v5 (python/openc3/io/json_api_object)"

    def __init__(
        self,
        url: str,
        timeout: float = 1.0,
        authentication: Optional[OpenC3Authentication] = None,
    ):
        """
        Args:
            url (str): The url of openc3-cosmos-cmd-tlm-api http://openc3-cosmos-cmd-tlm-api:2901
            timeout (float): The time to wait before disconnecting default = 1.0
            authentication (OpenC3Authentication): The authentication object if None initialize will generate default
        """
        self.http = None
        self.mutex = Lock()
        self.request_data = ""
        self.response_data = ""
        self.url: str = url
        self.log = [None, None, None]
        self.authentication = authentication if authentication else self.generate_auth()
        self.timeout: float = timeout
        self.shutdown: bool = False

    @staticmethod
    def generate_auth():
        """generate the auth object"""
        if OPENC3_API_TOKEN is None and OPENC3_API_USER is None:
            return OpenC3Authentication() if OPENC3_API_PASSWORD else None
        else:
            return OpenC3KeycloakAuthentication(OPENC3_KEYCLOAK_URL)

    def request(self, *method_params, **keyword_params):
        """Forwards all method calls to the remote service.
        Args:
            method_params (list) Array of parameters to pass to the method
            keyword_params (dict <Symbol, Variable>) Hash of keyword parameters
        Returns:
            return The result of the method call.
        """
        if self.shutdown:
            raise JsonApiError("Shutdown")
        method = method_params[0]
        endpoint = method_params[1]
        with self.mutex:
            kwargs = self._generate_kwargs(keyword_params)
            self.log = [None, None, None]
            if not self.http:
                self.connect()
            return self._send_request(method, endpoint, kwargs)

    def disconnect(self):
        """Disconnects from http server"""
        if self.http:
            self.http.close()
        self.http = None

    def shutdown(self):
        """Permanently disconnects from the http server"""
        self.shutdown = True
        self.disconnect()

    def connect(self):
        try:
            self.http = Session()
        except Exception as error:
            raise JsonApiError(error)

    def _generate_kwargs(self, keyword_params):
        """NOTE: This is a helper method and should not be called directly"""
        kwargs = {}
        for key, value in keyword_params.items():
            kwargs[key] = value

        kwargs["scope"] = self._generate_scope(kwargs)
        kwargs["headers"] = self._generate_headers(kwargs)
        kwargs["data"] = self._generate_data(kwargs)
        kwargs["query"] = self._generate_query(kwargs)
        kwargs["params"] = kwargs["query"]
        del kwargs["query"]
        del kwargs["scope"]
        return kwargs

    @staticmethod
    def _generate_scope(kwargs):
        """NOTE: This is a helper method and should not be called directly"""
        scope = kwargs.get("scope", None)
        if not scope:
            raise JsonApiError(f"no scope keyword found: {kwargs}")
        elif not isinstance(scope, str):
            raise JsonApiError(f"incorrect type for keyword 'scope' MUST be String: {scope}")
        return scope

    def _generate_headers(self, kwargs):
        """NOTE: This is a helper method and should not be called directly"""
        headers = kwargs.get("headers", None)
        if not headers:
            headers = kwargs["headers"] = {}
        elif not isinstance(headers, dict):
            raise JsonApiError(f"incorrect type for keyword 'headers' MUST be Dictionary: {headers}")

        if "json" in kwargs and kwargs["json"]:
            headers["Content-Type"] = "application/json"
        token = kwargs.get("token", None)
        if self.authentication and not token:
            token = self.authentication.token()
        if token:
            headers["User-Agent"] = self.USER_AGENT
            headers["Authorization"] = token
        else:
            headers["User-Agent"] = self.USER_AGENT
        return headers

    @staticmethod
    def _generate_data(kwargs):
        """NOTE: This is a helper method and should not be called directly"""
        data = kwargs.get("data", None)
        if not data:
            data = kwargs["data"] = {}
        elif not isinstance(data, dict) and not isinstance(data, str):
            raise JsonApiError(f"incorrect type for keyword 'data' MUST be Dictionary or String: {data}")
        if "json" in kwargs and kwargs["json"]:
            return json.dumps(kwargs["data"])
        else:
            return kwargs["data"]

    @staticmethod
    def _generate_query(kwargs):
        """NOTE: This is a helper method and should not be called directly"""
        query = kwargs.get("query", None)
        if query is None:
            query = kwargs["query"] = {}
        elif not isinstance(query, dict):
            raise JsonApiError(f"incorrect type for keyword 'query' MUST be Dictionary: {query}")
        if "scope" in kwargs and kwargs["scope"]:
            kwargs["query"]["scope"] = kwargs["scope"]
        return kwargs["query"]

    def _send_request(self, method, endpoint, kwargs):
        """NOTE: This is a helper method and should not be called directly"""
        try:
            kwargs["url"] = f"{self.url}{endpoint}"
            self.log[0] = f"{method} Request: {kwargs}"
            resp = getattr(self.http, method)(**kwargs)
            self.log[1] = f"{method} Response: {resp.status_code} {resp.headers} {resp.text}"
            self.response_data = resp.text
            return resp
        except Exception as error:
            self.log[2] = f"{method} Exception: {repr(error)}"
            self.disconnect()
            error = f"Api Exception: {self.log[0]} ::: {self.log[1]} ::: {self.log[2]}"
            raise RuntimeError(error)



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import os
import time
import json
from openc3.streams.web_socket_client_stream import WebSocketClientStream
from openc3.utilities.authentication import (
    OpenC3Authentication,
    OpenC3KeycloakAuthentication,
)
from openc3.utilities.time import to_nsec_from_epoch
from openc3.environment import OPENC3_SCOPE


# NOTE: For example usage see python/examples/cosmos_web_socket_example.py


# Base class - Do not use directly
class WebSocketApi:
    USER_AGENT = "OpenC3 / v5 (ruby/openc3/lib/io/web_socket_api)"

    # Create the WebsocketApi object
    def __init__(
        self,
        url,
        write_timeout=10.0,
        read_timeout=10.0,
        connect_timeout=5.0,
        authentication=None,
        scope=OPENC3_SCOPE,
    ):
        self.scope = scope
        if authentication is None:
            self.authentication = self._generate_auth()
        else:
            self.authentication = authentication
        self.url = url
        self.write_timeout = write_timeout
        self.read_timeout = read_timeout
        self.connect_timeout = connect_timeout
        self.subscribed = False

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    # Read the next message without filtering / parsing
    def read_message(self):
        self.subscribe()
        return self.stream.read()

    # Read the next message with json parsing, filtering, and timeout support
    def read(self, ignore_protocol_messages=True, timeout=None):
        start_time = time.time()
        while True:
            message = self.read_message()
            if message:
                json_hash = json.loads(message)
                if ignore_protocol_messages:
                    msg_type = json_hash.get("type")
                    if msg_type:  # ping, welcome, confirm_subscription, reject_subscription, disconnect
                        if msg_type == "disconnect":
                            if json_hash["reason"] == "unauthorized":
                                raise RuntimeError("Unauthorized")
                        if msg_type == "reject_subscription":
                            raise RuntimeError("Subscription Rejected")
                        if timeout:
                            end_time = time.time()
                            if (start_time - end_time) > timeout:
                                raise TimeoutError("No Data Timeout")
                        # if defined? RunningScript and RunningScript.instance:
                        #   if RunningScript.instance.stop?:
                        #       raise StopScript
                        continue
                return json_hash["message"]
            return message

    # Will subscribe to the channel based on @identifier
    def subscribe(self):
        if not self.subscribed:
            json_hash = {}
            json_hash["command"] = "subscribe"
            json_hash["identifier"] = json.dumps(self.identifier)
            self.stream.write(json.dumps(json_hash))
            self.subscribed = True

    # Will unsubscribe to the channel based on @identifier
    def unsubscribe(self):
        if self.subscribed:
            json_hash = {}
            json_hash["command"] = "unsubscribe"
            json_hash["identifier"] = json.dumps(self.identifier)
            self.stream.write(json.dumps(json_hash))
            self.subscribed = False

    # Send an ActionCable command
    def write_action(self, data_hash):
        json_hash = {}
        json_hash["command"] = "message"
        json_hash["identifier"] = json.dumps(self.identifier)
        json_hash["data"] = json.dumps(data_hash)
        self.write(json.dumps(json_hash))

    # General write to the websocket
    def write(self, data):
        self.subscribe()
        self.stream.write(data)

    # Connect to the websocket with authorization in query params
    def connect(self):
        self.disconnect()
        # Add the token directly in the URL since adding it to the header doesn't seem to work
        # Note in the this case we remove the "Bearer " string which is part of the token
        final_url = self.url + f"?scope={self.scope}&authorization={self.authentication.token(include_bearer=False)}"
        self.stream = WebSocketClientStream(final_url, self.write_timeout, self.read_timeout, self.connect_timeout)
        self.stream.headers = {
            "Sec-WebSocket-Protocol": "actioncable-v1-json, actioncable-unsupported",
            "User-Agent": WebSocketApi.USER_AGENT,
            # Adding the authorization token to the header is supposed to work
            # We add it directly with "Bearer <token>"
            # But for some reason it doesn't so we add it directly to the URL above
            # "Authorization": self.authentication.token(include_bearer=False),
        }
        return self.stream.connect()

    # Are we connected?
    def connected(self):
        if hasattr(self, "stream"):
            return self.stream.connected()
        else:
            return False

    # Disconnect from the websocket and attempt to send unsubscribe message
    def disconnect(self):
        if self.connected():
            self.unsubscribe()
            self.stream.disconnect()

    # Generate the appropriate token for OpenC3
    def _generate_auth(self):
        if os.environ.get("OPENC3_API_TOKEN") is None and os.environ.get("OPENC3_API_USER") is None:
            if os.environ.get("OPENC3_API_PASSWORD"):
                return OpenC3Authentication()
            else:
                return None
        else:
            return OpenC3KeycloakAuthentication(os.environ.get("OPENC3_KEYCLOAK_URL"))


# Base class for cmd-tlm-api websockets - Do not use directly
class CmdTlmWebSocketApi(WebSocketApi):
    def __init__(
        self,
        url=None,
        write_timeout=10.0,
        read_timeout=10.0,
        connect_timeout=5.0,
        authentication=None,
        scope=OPENC3_SCOPE,
    ):
        if not url:
            url = self.generate_url()
        super().__init__(
            url=url,
            write_timeout=write_timeout,
            read_timeout=read_timeout,
            connect_timeout=connect_timeout,
            authentication=authentication,
            scope=scope,
        )

    def generate_url(self):
        schema = os.environ.get("OPENC3_API_SCHEMA") or "http"
        if schema == "http":
            schema = "ws"
        if schema == "https":
            schema = "wss"
        hostname = os.environ.get("OPENC3_API_HOSTNAME") or (
            "127.0.0.1" if os.environ.get("OPENC3_DEVEL") else "openc3-cosmos-cmd-tlm-api"
        )
        port = os.environ.get("OPENC3_API_CABLE_PORT") or os.environ.get("OPENC3_API_PORT") or "3901"
        port = int(port)
        return f"{schema}://{hostname}:{port}/openc3-api/cable"


# Base class for script-runner-api websockets - Do not use directly
class ScriptWebSocketApi(WebSocketApi):
    def __init__(
        self,
        url=None,
        write_timeout=10.0,
        read_timeout=10.0,
        connect_timeout=5.0,
        authentication=None,
        scope=OPENC3_SCOPE,
    ):
        if not url:
            url = self.generate_url()
        super().__init__(
            url=url,
            write_timeout=write_timeout,
            read_timeout=read_timeout,
            connect_timeout=connect_timeout,
            authentication=authentication,
            scope=scope,
        )

    def generate_url(self):
        schema = os.environ.get("OPENC3_SCRIPT_API_SCHEMA") or "http"
        if schema == "http":
            schema = "ws"
        if schema == "https":
            schema = "wss"
        hostname = os.environ.get("OPENC3_SCRIPT_API_HOSTNAME") or (
            "127.0.0.1" if os.environ.get("OPENC3_DEVEL") else "openc3-cosmos-script-runner-api"
        )
        port = os.environ.get("OPENC3_SCRIPT_API_CABLE_PORT") or os.environ.get("OPENC3_SCRIPT_API_PORT") or "3902"
        port = int(port)
        return f"{schema}://{hostname}:{port}/script-api/cable"


# Running Script WebSocket
class RunningScriptWebSocketApi(ScriptWebSocketApi):
    def __init__(
        self,
        id,
        url=None,
        write_timeout=10.0,
        read_timeout=10.0,
        connect_timeout=5.0,
        authentication=None,
        scope=OPENC3_SCOPE,
    ):
        self.identifier = {"channel": "RunningScriptChannel", "id": id}
        super().__init__(
            url=url,
            write_timeout=write_timeout,
            read_timeout=read_timeout,
            connect_timeout=connect_timeout,
            authentication=authentication,
            scope=scope,
        )


# All Scripts WebSocket
class AllScriptsWebSocketApi(ScriptWebSocketApi):
    def __init__(
        self,
        url=None,
        write_timeout=10.0,
        read_timeout=10.0,
        connect_timeout=5.0,
        authentication=None,
        scope=OPENC3_SCOPE,
    ):
        self.identifier = {"channel": "AllScriptsChannel"}
        super().__init__(
            url=url,
            write_timeout=write_timeout,
            read_timeout=read_timeout,
            connect_timeout=connect_timeout,
            authentication=authentication,
            scope=scope,
        )


# Log Messages WebSocket
class MessagesWebSocketApi(CmdTlmWebSocketApi):
    def __init__(
        self,
        history_count=0,
        start_time=None,
        end_time=None,
        level=None,
        types=None,
        url=None,
        write_timeout=10.0,
        read_timeout=10.0,
        connect_timeout=5.0,
        authentication=None,
        scope=OPENC3_SCOPE,
    ):
        self.identifier = {"channel": "MessagesChannel", "history_count": history_count}
        if start_time:
            self.identifier["start_time"] = start_time
        if end_time:
            self.identifier["end_time"] = end_time
        if level:
            self.identifier["level"] = level
        if types:
            self.identifier["types"] = types
        super().__init__(
            url=url,
            write_timeout=write_timeout,
            read_timeout=read_timeout,
            connect_timeout=connect_timeout,
            authentication=authentication,
            scope=scope,
        )


# Autonomic Events WebSocket (Enterprise Only)
class AutonomicEventsWebSocketApi(CmdTlmWebSocketApi):
    def __init__(
        self,
        history_count=0,
        url=None,
        write_timeout=10.0,
        read_timeout=10.0,
        connect_timeout=5.0,
        authentication=None,
        scope=OPENC3_SCOPE,
    ):
        self.identifier = {
            "channel": "AutonomicEventsChannel",
            "history_count": history_count,
        }
        super().__init__(
            url=url,
            write_timeout=write_timeout,
            read_timeout=read_timeout,
            connect_timeout=connect_timeout,
            authentication=authentication,
            scope=scope,
        )


# Calendar Events WebSocket (Enterprise Only)
class CalendarEventsWebSocketApi(CmdTlmWebSocketApi):
    def __init__(
        self,
        history_count=0,
        url=None,
        write_timeout=10.0,
        read_timeout=10.0,
        connect_timeout=5.0,
        authentication=None,
        scope=OPENC3_SCOPE,
    ):
        self.identifier = {
            "channel": "CalendarEventsChannel",
            "history_count": history_count,
        }
        super().__init__(
            url=url,
            write_timeout=write_timeout,
            read_timeout=read_timeout,
            connect_timeout=connect_timeout,
            authentication=authentication,
            scope=scope,
        )


# Config Events WebSocket
class ConfigEventsWebSocketApi(CmdTlmWebSocketApi):
    def __init__(
        self,
        history_count=0,
        url=None,
        write_timeout=10.0,
        read_timeout=10.0,
        connect_timeout=5.0,
        authentication=None,
        scope=OPENC3_SCOPE,
    ):
        self.identifier = {
            "channel": "ConfigEventsChannel",
            "history_count": history_count,
        }
        super().__init__(
            url=url,
            write_timeout=write_timeout,
            read_timeout=read_timeout,
            connect_timeout=connect_timeout,
            authentication=authentication,
            scope=scope,
        )


# Limits Events WebSocket
class LimitsEventsWebSocketApi(CmdTlmWebSocketApi):
    def __init__(
        self,
        history_count=0,
        url=None,
        write_timeout=10.0,
        read_timeout=10.0,
        connect_timeout=5.0,
        authentication=None,
        scope=OPENC3_SCOPE,
    ):
        self.identifier = {
            "channel": "LimitsEventsChannel",
            "history_count": history_count,
        }
        super().__init__(
            url=url,
            write_timeout=write_timeout,
            read_timeout=read_timeout,
            connect_timeout=connect_timeout,
            authentication=authentication,
            scope=scope,
        )


# Timeline WebSocket
class TimelineEventsWebSocketApi(CmdTlmWebSocketApi):
    def __init__(
        self,
        history_count=0,
        url=None,
        write_timeout=10.0,
        read_timeout=10.0,
        connect_timeout=5.0,
        authentication=None,
        scope=OPENC3_SCOPE,
    ):
        self.identifier = {
            "channel": "TimelineEventsChannel",
            "history_count": history_count,
        }
        super().__init__(
            url=url,
            write_timeout=write_timeout,
            read_timeout=read_timeout,
            connect_timeout=connect_timeout,
            authentication=authentication,
            scope=scope,
        )


# Streaming API WebSocket
class StreamingWebSocketApi(CmdTlmWebSocketApi):
    def __init__(
        self,
        url=None,
        write_timeout=10.0,
        read_timeout=10.0,
        connect_timeout=5.0,
        authentication=None,
        scope=OPENC3_SCOPE,
    ):
        self.identifier = {"channel": "StreamingChannel"}
        super().__init__(
            url=url,
            write_timeout=write_timeout,
            read_timeout=read_timeout,
            connect_timeout=connect_timeout,
            authentication=authentication,
            scope=scope,
        )

    # Request to add data to the stream
    #
    # arguments:
    # scope: scope name
    # start_time: 64-bit nanoseconds from unix epoch - If not present then realtime
    # end_time: 64-bit nanoseconds from unix epoch - If not present stream forever
    # items: [ [ MODE__CMDORTLM__TARGET__PACKET__ITEM__VALUETYPE__REDUCEDTYPE, item_key] ]
    #   MODE - RAW, DECOM, REDUCED_MINUTE, REDUCED_HOUR, or REDUCED_DAY
    #   CMDORTLM - CMD or TLM
    #   TARGET - Target name
    #   PACKET - Packet name
    #   ITEM - Item Name
    #   VALUETYPE - RAW, CONVERTED, FORMATTED, or WITH_UNITS
    #   REDUCEDTYPE - MIN, MAX, AVG, STDDEV (only for reduced modes)
    #   item_key is an optional shortened name to return the data as
    # packets: [ MODE__CMDORTLM__TARGET__PACKET__VALUETYPE ]
    #   MODE - RAW, DECOM, REDUCED_MINUTE, REDUCED_HOUR, or REDUCED_DAY
    #   CMDORTLM - CMD or TLM
    #   TARGET - Target name
    #   PACKET - Packet name
    #   VALUETYPE - RAW, CONVERTED, FORMATTED, WITH_UNITS, or PURE (pure means all types as stored in log)
    #
    def add(
        self,
        items=None,
        packets=None,
        start_time=None,
        end_time=None,
        scope=OPENC3_SCOPE,
    ):
        data_hash = {}
        data_hash["action"] = "add"
        if start_time:
            data_hash["start_time"] = to_nsec_from_epoch(start_time)
        if end_time:
            data_hash["end_time"] = to_nsec_from_epoch(end_time)
        if items:
            data_hash["items"] = items
        if packets:
            data_hash["packets"] = packets
        data_hash["scope"] = scope
        data_hash["token"] = self.authentication.token(include_bearer=False)
        self.write_action(data_hash)

    # Request to remove data from the stream
    #
    # arguments:
    # scope: scope name
    # items: [ [ MODE__CMDORTLM__TARGET__PACKET__ITEM__VALUETYPE__REDUCEDTYPE] ]
    #   MODE - RAW, DECOM, REDUCED_MINUTE, REDUCED_HOUR, or REDUCED_DAY
    #   CMDORTLM - CMD or TLM
    #   TARGET - Target name
    #   PACKET - Packet name
    #   ITEM - Item Name
    #   VALUETYPE - RAW, CONVERTED, FORMATTED, or WITH_UNITS
    #   REDUCEDTYPE - MIN, MAX, AVG, STDDEV (only for reduced modes)
    # packets: [ MODE__CMDORTLM__TARGET__PACKET__VALUETYPE ]
    #   MODE - RAW, DECOM, REDUCED_MINUTE, REDUCED_HOUR, or REDUCED_DAY
    #   CMDORTLM - CMD or TLM
    #   TARGET - Target name
    #   PACKET - Packet name
    #   VALUETYPE - RAW, CONVERTED, FORMATTED, WITH_UNITS, or PURE (pure means all types as stored in log)
    #
    def remove(self, items=None, packets=None, scope=OPENC3_SCOPE):
        data_hash = {}
        data_hash["action"] = "remove"
        if items:
            data_hash["items"] = items
        if packets:
            data_hash["packets"] = packets
        data_hash["scope"] = scope
        data_hash["token"] = self.authentication.token(include_bearer=False)
        self.write_action(data_hash)

    # Convenience method to read all data until end marker is received.
    # Warning: DATA IS STORED IN RAM.  Do not use this with large queries
    @classmethod
    def read_all(
        cls,
        items=None,
        packets=None,
        start_time=None,
        end_time=None,
        scope=OPENC3_SCOPE,
        timeout=None,
    ):
        read_all_start_time = time.time()
        data = []
        with cls() as api:
            api.add(
                items=items,
                packets=packets,
                start_time=start_time,
                end_time=end_time,
                scope=scope,
            )
            while True:
                batch = api.read()
                if len(batch) == 0:
                    break
                else:
                    data += batch
                if timeout and (time.time() - read_all_start_time) > timeout:
                    break
        return data



# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

# Modified by OpenC3, Inc.
# All changes Copyright 2023, OpenC3, Inc.
# All Rights Reserved
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import functools
import requests


def request_wrapper(func):
    @functools.wraps(func)
    def _request(*args, **kwargs):
        try:
            value = func(*args, **kwargs)
            return value
        except ValueError as exc:
            err = f"ValueError {exc} while requesting token."
            raise RuntimeError(err) from exc
        except requests.Timeout as exc:
            err = f"Timeout error while requesting {exc.request.url!r}"
            raise RuntimeError(err) from exc
        except requests.HTTPError as exc:
            err = f"Error response {exc.response.status_code} while requesting {exc.request.url!r}."
            if 400 >= exc.response.status_code:
                raise RuntimeError(err) from exc
        except requests.RequestException as exc:
            err = f"An error occurred while requesting {exc.request.url!r}."
            raise RuntimeError(err) from exc

    return _request



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from datetime import datetime, timezone
from openc3.system.system import System
from openc3.config.config_parser import ConfigParser
from openc3.interfaces.protocols.burst_protocol import BurstProtocol


# Delineates packets by identifying them and then
# reading out their entire fixed length. Packets lengths can vary but
# they must all be fixed.
class FixedProtocol(BurstProtocol):
    # self.param min_id_size [Integer] The minimum amount of data needed to
    #   identify a packet.
    # self.param discard_leading_bytes (see BurstProtocol#initialize)
    # self.param sync_pattern (see BurstProtocol#initialize)
    # self.param telemetry [Boolean] Whether the interface is returning
    #   telemetry (True) or commands (False)
    # self.param fill_fields (see BurstProtocol#initialize)
    # self.param unknown_raise Whether to raise an exception on an unknown packet
    # self.param allow_empty_data [True/False/None] See Protocol#initialize
    def __init__(
        self,
        min_id_size,
        discard_leading_bytes=0,
        sync_pattern=None,
        telemetry=True,
        fill_fields=False,
        unknown_raise=False,
        allow_empty_data=None,
    ):
        super().__init__(discard_leading_bytes, sync_pattern, fill_fields, allow_empty_data)
        self.min_id_size = int(min_id_size)
        self.telemetry = telemetry
        self.unknown_raise = ConfigParser.handle_true_false(unknown_raise)
        self.received_time = None
        self.target_name = None
        self.packet_name = None

    # Set the received_time, target_name and packet_name which we recorded when
    # we identified this packet. The server will also do this but since we know
    # the information here, we perform this optimization.
    def read_packet(self, packet):
        packet.received_time = self.received_time
        packet.target_name = self.target_name
        packet.packet_name = self.packet_name
        return packet

    # Identifies an unknown buffer of data as a Packet. The raw data is
    # returned but the packet that matched is recorded so it can be set in the
    # read_packet callback.
    #
    # self.return [String|Symbol] The identified packet data or 'STOP' if more data:
    #   is required to build a packet
    def identify_and_finish_packet(self):
        packet_data = None
        identified_packet = None

        if self.telemetry:
            target_names = self.interface.tlm_target_names
        else:
            target_names = self.interface.cmd_target_names

        for target_name in target_names:
            target_packets = None
            unique_id_mode = False
            try:
                if self.telemetry:
                    target_packets = System.telemetry.packets(target_name)
                    target = System.targets[target_name]
                    if target:
                        unique_id_mode = target.tlm_unique_id_mode
                else:
                    target_packets = System.commands.packets(target_name)
                    target = System.targets[target_name]
                    if target:
                        unique_id_mode = target.cmd_unique_id_mode
            except RuntimeError as error:
                if "does not exist" in repr(error):
                    # No commands/telemetry for this target
                    continue
                else:
                    raise error

            if unique_id_mode:
                for _, packet in target_packets.items():
                    if packet.identify(self.data[self.discard_leading_bytes :]):
                        identified_packet = packet
                        break
            else:
                # Do a lookup to quickly identify the packet
                if len(target_packets) > 0:
                    packet = next(iter(target_packets.values()))
                    key = packet.read_id_values(self.data[self.discard_leading_bytes :])
                    if self.telemetry:
                        id_values = System.telemetry.config.tlm_id_value_hash[target_name]
                    else:
                        id_values = System.commands.config.cmd_id_value_hash[target_name]
                    identified_packet = id_values.get(repr(key))
                    if identified_packet is None:
                        identified_packet = id_values.get("CATCHALL")

            if identified_packet is not None:
                if identified_packet.defined_length + self.discard_leading_bytes > len(self.data):
                    # Check if need more data to finish packet:
                    return ("STOP", self.extra)

                # Set some variables so we can update the packet in
                # read_packet
                self.received_time = datetime.now(timezone.utc)
                self.target_name = identified_packet.target_name
                self.packet_name = identified_packet.packet_name

                # Get the data from this packet
                packet_data = self.data[0 : (identified_packet.defined_length + self.discard_leading_bytes)]
                self.data = self.data[(identified_packet.defined_length + self.discard_leading_bytes) :]
                break

        if identified_packet is None:
            if self.unknown_raise:
                raise RuntimeError("Unknown data received by FixedProtocol")

            # Unknown packet? Just return all the current data
            self.received_time = None
            self.target_name = None
            self.packet_name = None
            packet_data = self.data[:]
            self.data = b""

        return (packet_data, self.extra)

    def reduce_to_single_packet(self):
        if len(self.data) < self.min_id_size:
            return ("STOP", self.extra)

        return self.identify_and_finish_packet()



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.accessors.xml_accessor import XmlAccessor
# XmlAccessor uses lxml.etree
import lxml.html as HTML


class HtmlAccessor(XmlAccessor):
    @classmethod
    def _buffer_to_doc(cls, buffer):
        # Override the XML implementation to use the HTML parser
        return HTML.fromstring(buffer.decode())

    @classmethod
    def _doc_to_buffer(cls, doc):
        return bytearray(HTML.tostring(doc))



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import math
import statistics
from openc3.processors.processor import Processor


class StatisticsProcessor(Processor):
    # @param item_name [String] The name of the item to gather statistics on
    # @param samples_to_average [Integer] The number of samples to store for calculations
    # @param value_type #See Processor::initialize
    def __init__(self, item_name, samples_to_average, value_type="CONVERTED"):
        super().__init__(value_type)
        self.item_name = str(item_name).upper()
        self.samples_to_average = int(samples_to_average)
        self.reset()

    # Run statistics on the item
    #
    # See Processor#call
    def call(self, packet, buffer):
        value = packet.read(self.item_name, self.value_type, buffer)
        # Don't process NaN or Infinite values
        if math.isnan(value) or math.isinf(value):
            return

        self.samples.append(value)
        if len(self.samples) > self.samples_to_average:
            self.samples = self.samples[-self.samples_to_average :]
        self.results["MAX"] = max(self.samples)
        self.results["MIN"] = min(self.samples)
        self.results["MEAN"] = statistics.fmean(self.samples)
        if len(self.samples) > 1:
            self.results["STDDEV"] = statistics.stdev(self.samples)
        else:
            self.results["STDDEV"] = 0

    # Reset any state
    def reset(self):
        self.samples = []
        self.results["MAX"] = None
        self.results["MIN"] = None
        self.results["MEAN"] = None
        self.results["STDDEV"] = None

    # Make a light weight clone of this processor. This only creates a new hash of results
    #
    # @return [Processor] A copy of the processor with a new hash of results
    def clone(self):
        processor = super().clone()
        processor.samples = processor.samples[:]
        return processor

    # Convert to configuration file string
    def to_config(self):
        return f"  PROCESSOR {self.name} {self.__class__.__name__} {self.item_name} {self.samples_to_average} {self.value_type}\n"

    def as_json(self):
        return {
            "name": self.name,
            "class": self.__class__.__name__,
            "params": [self.item_name, self.samples_to_average, self.value_type],
        }



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.
#
# A portion of this file was funded by Blue Origin Enterprises, L.P.
# See https://github.com/OpenC3/cosmos/pull/1963

from datetime import datetime, timezone
import json
from openc3.system.system import System
from openc3.topics.topic import Topic
from openc3.topics.telemetry_topic import TelemetryTopic
from openc3.utilities.time import to_nsec_from_epoch
from openc3.utilities.json import JsonEncoder, JsonDecoder
from openc3.models.target_model import TargetModel

def handle_inject_tlm(inject_tlm_json, scope):
    inject_tlm_hash = json.loads(inject_tlm_json, cls=JsonDecoder)
    target_name = inject_tlm_hash["target_name"]
    packet_name = inject_tlm_hash["packet_name"]
    item_hash = inject_tlm_hash["item_hash"]
    type = str(inject_tlm_hash["type"])
    packet = System.telemetry.packet(target_name, packet_name)
    if item_hash:
        for name, value in item_hash.items():
            packet.write(str(name), value, type)
    packet.received_time = datetime.now(timezone.utc)
    packet.received_count = TargetModel.increment_telemetry_count(packet.target_name, packet.packet_name, 1, scope=scope)
    TelemetryTopic.write_packet(packet, scope)


def handle_build_cmd(build_cmd_json, msg_id, scope):
    build_cmd_hash = json.loads(build_cmd_json, cls=JsonDecoder)
    target_name = build_cmd_hash["target_name"]
    cmd_name = build_cmd_hash["cmd_name"]
    cmd_params = build_cmd_hash["cmd_params"]
    range_check = build_cmd_hash["range_check"]
    raw = build_cmd_hash["raw"]
    ack_topic = f"{{{scope}__ACKCMD}}TARGET__{target_name}"
    try:
        command = System.commands.build_cmd(target_name, cmd_name, cmd_params, range_check, raw)
        msg_hash = {
            "id": msg_id,
            "result": "SUCCESS",
            "time": to_nsec_from_epoch(command.packet_time),
            "received_time": to_nsec_from_epoch(command.received_time),
            "target_name": command.target_name,
            "packet_name": command.packet_name,
            "received_count": command.received_count,
            "buffer": json.dumps(command.buffer_no_copy(), cls=JsonEncoder),
        }
    # If there is an error due to parameter out of range, etc, we rescue it so we can
    # write the ACKCMD}TARGET topic and allow the TelemetryDecomTopic.build_cmd to return
    except RuntimeError as error:
        msg_hash = {"id": msg_id, "result": "ERROR", "message": repr(error)}
    Topic.write_topic(ack_topic, msg_hash)



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.
#
# A portion of this file was funded by Blue Origin Enterprises, L.P.
# See https://github.com/OpenC3/cosmos/pull/1953

from typing import Optional

from openc3.environment import OPENC3_SCOPE
from openc3.models.model import Model

# require 'openc3/models/metric_model'
from openc3.utilities.bucket import Bucket


class MicroserviceModel(Model):
    PRIMARY_KEY = "openc3_microservices"

    # NOTE: The following three class methods are used by the ModelController
    # and are reimplemented to enable various Model class methods to work
    @classmethod
    def get(cls, name, scope: Optional[str] = None):
        return super().get(MicroserviceModel.PRIMARY_KEY, name)

    @classmethod
    def names(cls, scope: Optional[str] = None):
        scoped = []
        unscoped = super().names(MicroserviceModel.PRIMARY_KEY)
        for name in unscoped:
            if scope is None or name.split("__")[0] == scope:
                scoped.append(name)
        return scoped

    @classmethod
    def all(cls, scope: Optional[str] = None):
        scoped = {}
        unscoped = super().all(MicroserviceModel.PRIMARY_KEY)
        for name, json in unscoped.items():
            if scope is None or name.split("__")[0] == scope:
                scoped[name] = json
        return scoped

    # Create a microservice model to be deployed to bucket storage
    def __init__(
        self,
        name: str,
        folder_name: str = None,
        cmd: Optional[list] = None,
        work_dir: str = ".",
        ports: Optional[list] = None,
        env: Optional[dict] = None,
        topics: Optional[list] = None,
        target_names: Optional[list] = None,
        options: Optional[list] = None,
        parent=None,
        container=None,
        updated_at: Optional[float] = None,
        plugin=None,
        needs_dependencies=False,
        secrets: Optional[list] = None,
        prefix=None,
        disable_erb=None,
        ignore_changes=None,
        shard=0,
        enabled: bool = True,
        scope: str = OPENC3_SCOPE,
    ):
        parts = name.split("__")
        if len(parts) != 3:
            raise RuntimeError(f"name '{name}' must be formatted as SCOPE__TYPE__NAME")
        if parts[0] != scope:
            raise RuntimeError(f"name '{name}' scope '{parts[0]}' doesn't match scope parameter '{scope}'")

        cmd = [] if cmd is None else cmd
        ports = [] if ports is None else ports
        env = {} if env is None else env
        topics = [] if topics is None else topics
        target_names = [] if target_names is None else target_names
        options = [] if options is None else options
        secrets = [] if secrets is None else secrets

        super().__init__(
            MicroserviceModel.PRIMARY_KEY,
            name=name,
            updated_at=updated_at,
            plugin=plugin,
            scope=scope,
        )
        self.folder_name = folder_name
        self.cmd = cmd
        self.work_dir = work_dir
        self.ports = ports
        self.env = env
        self.topics = topics
        self.target_names = target_names
        self.options = options
        self.parent = parent
        self.container = container
        self.needs_dependencies = needs_dependencies
        self.secrets = secrets
        self.prefix = prefix
        self.disable_erb = disable_erb
        self.ignore_changes = ignore_changes
        self.shard = shard
        if self.shard is None:
            self.shard = 0
        self.enabled = enabled
        if self.enabled is None:
            self.enabled = True
        self.bucket = Bucket.getClient()

    def as_json(self):
        return {
            "name": self.name,
            "folder_name": self.folder_name,
            "cmd": self.cmd,
            "work_dir": self.work_dir,
            "ports": self.ports,
            "env": self.env,
            "topics": self.topics,
            "target_names": self.target_names,
            "options": self.options,
            "parent": self.parent,
            "container": self.container,
            "updated_at": self.updated_at,
            "plugin": self.plugin,
            "needs_dependencies": self.needs_dependencies,
            "secrets": self.secrets,  # .as_json(),
            "prefix": self.prefix,
            "disable_erb": self.disable_erb,
            "ignore_changes": self.ignore_changes,
            "shard": self.shard,
            "enabled": self.enabled,
        }



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import os
import time
from datetime import datetime, timezone
from threading import Lock
from openc3.utilities.bucket_utilities import BucketUtilities
from openc3.utilities.logger import Logger
from openc3.environment import OPENC3_SCOPE


# Handles writing message logs to a file
class MessageLog:
    @classmethod
    def build_timestamped_filename(cls, tags=None, extension=".txt", time=None):
        if time is None:
            time = datetime.now(timezone.utc)
        timestamp = time.strftime("%Y_%m_%d_%H_%M_%S")
        tags = tags or []
        tags = [i for i in tags if i is not None]  # Remove Nones
        combined_tags = "_".join(tags)
        filename = None
        if len(combined_tags) > 0:
            filename = timestamp + "_" + combined_tags + extension
        else:
            filename = timestamp + extension
        return filename

    # @param tool_name [String] The name of the tool creating the message log.
    #   This will be inserted into the message log filename to help identify it.
    # @param log_dir [String] The filesystem path to store the message log file.
    # @param tags [Array<String>] Array of strings to put into the filename
    def __init__(self, tool_name, log_dir, tags=["messages"], scope=OPENC3_SCOPE):
        self.remote_log_directory = f"{scope}/tool_logs/{tool_name}/"
        self.tags = [tool_name] + tags
        self.log_dir = log_dir
        self.filename = ""
        self.file = None
        self.start_day = None
        self.mutex = Lock()

    # Ensures the log file is opened and ready to write. It then writes the
    # message to the log and flushes it to force the write.
    #
    # @param message [String] Message to write to the log
    def write(self, message, flush=False):
        with self.mutex:
            if self.file is None or self.file.closed or not os.path.exists(self.filename):
                self.start(False)

            self.file.write(message)
            if flush:
                self.file.flush()

    # Closes the message log and marks it read only
    def stop(self, take_mutex=True, metadata={}):
        bucket_key = None
        if take_mutex:
            self.mutex.acquire()
        if self.file and not self.file.closed:
            self.file.close()
            os.chmod(self.filename, 0o444)
            bucket_key = os.path.join(
                self.remote_log_directory,
                self.start_day,
                os.path.basename(self.filename),
            )
            try:
                thread = BucketUtilities.move_log_file_to_bucket(self.filename, bucket_key, metadata=metadata)
                thread.join()
            except Exception as e:
                Logger.error(str(e))

        if take_mutex:
            self.mutex.release()
        return bucket_key

    # Creates a new message log and sets the filename
    def start(self, take_mutex=True):
        if take_mutex:
            self.mutex.acquire()
        # Prevent starting files too fast
        while True:
            if os.path.exists(os.path.join(self.log_dir, self.build_timestamped_filename(self.tags))):
                time.sleep(0.1)
            else:
                break
        self.stop(False)
        timed_filename = self.build_timestamped_filename(self.tags)
        self.start_day = timed_filename[0:10].replace("_", "")  # YYYYMMDD
        self.filename = os.path.join(self.log_dir, timed_filename)
        self.file = open(self.filename, "a")
        if take_mutex:
            self.mutex.release()



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
import time
import datetime
from typing import Optional

from openc3.utilities.store import Store, EphemeralStore
from openc3.utilities.store_queued import StoreQueued, EphemeralStoreQueued


class Model:
    @classmethod
    def store(cls):
        return Store

    @classmethod
    def store_queued(cls):
        return StoreQueued

    # NOTE: The following three methods must be reimplemented by Model subclasses
    # without primary_key to support other class methods.

    @classmethod
    def get(cls, primary_key: str, name: str):
        """
        Return:
            (Dict[] | None) Hash of this model or nil if name not found under primary_key
        """
        json_data = cls.store().hget(primary_key, name)
        return json.loads(json_data) if json_data else None

    @classmethod
    def names(cls, primary_key: str):
        """
        Return:
            (List[str]) All the names stored under the primary key
        """
        keys = cls.store().hkeys(primary_key)
        keys.sort()
        return [key.decode() for key in keys]

    @classmethod
    def all(cls, primary_key: str):
        """
        Return:
             (List[Dict]) All the models (as Hash objects) stored under the primary key
        """
        base = cls.store().hgetall(primary_key)
        # decode the binary string keys to strings
        decoded = {k.decode(): v for (k, v) in base.items()}
        for key, value in decoded.items():
            decoded[key] = json.loads(value)
        return decoded

    # END NOTE

    @classmethod
    def set(cls, json_data: dict, scope: str, queued: bool = False):
        """Sets (updates) the redis hash of this model"""
        json_data["scope"] = scope
        cls(**json_data).create(force=True, queued=queued)

    @classmethod
    def from_json(cls, json_data: str | dict, scope: str):
        """
        Return:
            [Model] Model generated from the passed JSON
        """
        if isinstance(json_data, str):
            json_data = json.loads(json_data)
        if json_data is None:
            raise RuntimeError("json data is nil")
        json_data["scope"] = scope
        return cls(**json_data)

    @classmethod
    def get_model(cls, name: str, scope: str):
        """Calls self.get_model and then from_json to turn the Hash configuration into a Ruby Model object.
        Return:
            [Object|nil] Model object or nil if name not found under primary_key
        """
        json_data = cls.get(name, scope)
        return cls.from_json(json_data, scope) if json_data else None

    # NOTE: get_all_models not implemented as it is currently
    # unused by any python models

    # NOTE: find_all_by_plugin, handle_config not implemented as it is
    # only needed by plugin_model which is Ruby only

    def __init__(self, primary_key: str, **kw_args):
        """Store the primary key and keyword arguments"""
        self.primary_key: str = primary_key
        self.name: Optional[str] = kw_args.get("name")
        self.updated_at: Optional[float] = kw_args.get("updated_at")
        self.plugin: Optional[str] = kw_args.get("plugin")
        self.scope: Optional[str] = kw_args.get("scope")
        self.destroyed: bool = False

    def create(self, update=False, force=False, queued=False, isoformat=False):
        """Update the Redis hash at primary_key and set the field "name"
        to the JSON generated via calling as_json
        """
        if not force:
            existing = self.store().hget(self.primary_key, self.name)
            if existing and not update:
                raise RuntimeError(f"{self.primary_key}:{self.name} already exists at create")
            if not existing and update:
                raise RuntimeError(f"{self.primary_key}:{self.name} doesn't exist at update")
        if isoformat:
            self.updated_at = datetime.datetime.now().isoformat()
        else:
            self.updated_at = time.time() * 1_000_000_000

        write_store = self.store_queued() if queued else self.store()
        write_store.hset(self.primary_key, self.name, json.dumps(self.as_json()))

    def update(self, force=False, queued=True):
        """Alias for create(update: true)"""
        self.create(update=True, force=force, queued=queued)

    def deploy(self, gem_path: str, variables: str):
        """Deploy the model into the OpenC3 system. Subclasses must implement this
        and typically create MicroserviceModels to implement.
        """
        raise NotImplementedError("must be implemented by subclass")

    def undeploy(self):
        """Undo the actions of deploy and remove the model from OpenC3.
        Subclasses must implement this as by default it is a noop.
        """
        pass

    def destroy(self):
        """Delete the model from the Store"""
        self.destroyed = True
        self.undeploy()
        self.store().hdel(self.primary_key, self.name)

    def as_json(self):
        """
        Return:
            (dict) JSON encoding of this model
        """
        return {
            "name": self.name,
            "updated_at": self.updated_at,
            "plugin": self.plugin,
            "scope": self.scope,
        }


class EphemeralModel(Model):
    @classmethod
    def store(cls):
        return EphemeralStore

    @classmethod
    def store_queued(cls):
        return EphemeralStoreQueued



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
import datetime


class JsonEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime.datetime):
            return o.strftime("%Y-%m-%d %H:%M:%S.%f")
        if isinstance(o, (bytes, bytearray)):
            return {
                "json_class": "String",
                "raw": [x for x in o],
            }
        return json.JSONEncoder.default(self, o)


class JsonDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, dct):
        if dct.get("json_class") == "String":
            return bytes(dct["raw"])
        return dct



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.environment import OPENC3_SCOPE
from openc3.models.model import Model


class MicroserviceStatusModel(Model):
    PRIMARY_KEY = "openc3_microservice_status"

    # NOTE: The following three class methods are used by the ModelController
    # and are reimplemented to enable various Model class methods to work
    @classmethod
    def get(cls, name: str, scope: str = OPENC3_SCOPE):
        return super().get(f"{scope}__{MicroserviceStatusModel.PRIMARY_KEY}", name=name)

    @classmethod
    def names(cls, scope: str = OPENC3_SCOPE):
        return super().names(f"{scope}__{MicroserviceStatusModel.PRIMARY_KEY}")

    @classmethod
    def all(cls, scope: str = OPENC3_SCOPE):
        return super().all(f"{scope}__{MicroserviceStatusModel.PRIMARY_KEY}")

    def __init__(
        self,
        name,
        state=None,
        count=0,
        error=None,
        custom=None,
        updated_at=None,
        plugin=None,
        scope=None,
    ):
        super().__init__(
            f"{scope}__{MicroserviceStatusModel.PRIMARY_KEY}",
            name=name,
            updated_at=updated_at,
            plugin=plugin,
            scope=scope,
        )
        self.state = state
        self.count = count
        self.error = error
        self.custom = custom

    def as_json(self):
        json = {
            "name": self.name,
            "state": self.state,
            "count": self.count,
            "plugin": self.plugin,
            "updated_at": self.updated_at,
        }
        if self.error is not None:
            json["error"] = repr(self.error)
        if self.custom is not None:
            json["custom"] = self.custom.as_json()
        return json
